# sage.doctest: optional - macaulay2
r"""
Interface to Macaulay2

.. NOTE::

    You must have ``Macaulay2`` installed on your computer
    for this interface to work. Macaulay2 is not included with Sage,
    but you can obtain it from https://macaulay2.com/.
    No additional optional Sage packages are required.

Sage provides an interface to the Macaulay2 computational algebra
system. This system provides extensive functionality for commutative
algebra. You do not have to install any optional packages.

The Macaulay2 interface offers three pieces of functionality:

- ``macaulay2_console()`` -- a function that dumps you
  into an interactive command-line Macaulay2 session

- ``macaulay2.eval(expr)`` -- evaluation of arbitrary Macaulay2
  expressions, with the result returned as a string

- ``macaulay2(expr)`` -- creation of a Sage object that wraps a
  Macaulay2 object.  This provides a Pythonic interface to Macaulay2.  For
  example, if ``f = macaulay2(10)``, then ``f.gcd(25)`` returns the
  GCD of `10` and `25` computed using Macaulay2.

EXAMPLES::

    sage: macaulay2('3/5 + 7/11')
    68
    --
    55
    sage: f = macaulay2('f = i -> i^3')
    sage: f
    f
    sage: f(5)
    125

    sage: R = macaulay2('ZZ/5[x,y,z]')
    sage: R
    ZZ
    --[x...z]
     5
    sage: x = macaulay2('x')
    sage: y = macaulay2('y')
    sage: (x+y)^5
     5    5
    x  + y
    sage: parent((x+y)^5)
    Macaulay2

The name of the variable to which a Macaulay2 element is assigned internally
can be passed as an argument. This is useful for types like polynomial rings
which acquire that name in Macaulay2::

    sage: R = macaulay2('QQ[x,y,z,w]', 'R')
    sage: R
    R

    sage: f = macaulay2('x^4 + 2*x*y^3 + x*y^2*w + x*y*z*w + x*y*w^2'
    ....:               '+ 2*x*z*w^2 + y^4 + y^3*w + 2*y^2*z*w + z^4 + w^4')
    sage: f
     4       3    4    4      2     3                2           2         2    4
    x  + 2x*y  + y  + z  + x*y w + y w + x*y*z*w + 2y z*w + x*y*w  + 2x*z*w  + w
    sage: g = f * macaulay2('x+y^5')
    sage: print(g.factor())
      4       3    4    4      2     3                2           2         2    4   5
    (x  + 2x*y  + y  + z  + x*y w + y w + x*y*z*w + 2y z*w + x*y*w  + 2x*z*w  + w )(y  + x)

Use :meth:`eval` for explicit control over what is sent to the interpreter.
The argument is evaluated in Macaulay2 as is::

    sage: macaulay2.eval('compactMatrixForm')
    true
    sage: macaulay2.eval('compactMatrixForm = false;')
    sage: macaulay2.eval('matrix {{1, x^2+y}}')
    |      2      |
    |  1  x  + y  |
    <BLANKLINE>
            1      2
    Matrix R  <-- R
    sage: macaulay2.eval('compactMatrixForm = true;')


AUTHORS:

- Kiran Kedlaya and David Roe (2006-02-05, during Sage coding sprint)
- William Stein (2006-02-09): inclusion in Sage; prompt uses regexp,
  calling of Macaulay2 functions via __call__.
- William Stein (2006-02-09): fixed bug in reading from file and
  improved output cleaning.
- Kiran Kedlaya (2006-02-12): added ring and ideal constructors,
  list delimiters, is_Macaulay2Element, sage_polystring,
  __floordiv__, __mod__, __iter__, __len__; stripped extra
  leading space and trailing newline from output.

.. TODO::

    Get rid of all numbers in output, e.g., in ideal function below.
"""

# ****************************************************************************
#       Copyright (C) 2006 Kiran S. Kedlaya <kedlaya@mit.edu>
#                          David Roe <roed@mit.edu>
#                          William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import os
import re
from collections.abc import Iterator

import sage.interfaces.abc
from sage.interfaces.expect import (
    Expect,
    ExpectElement,
    ExpectFunction,
    FunctionElement,
)
from sage.interfaces.interface import AsciiArtString
from sage.interfaces.tab_completion import ExtraTabCompletion
from sage.misc.instancedoc import instancedoc
from sage.misc.multireplace import multiple_replace
from sage.structure.global_options import GlobalOptions


def remove_output_labels(s) -> str:
    r"""
    Remove output labels of Macaulay2 from a string.

    - s: output of Macaulay2

    - s: string

    Returns: the input string with `n` symbols removed from the beginning of
    each line, where `n` is the minimal number of spaces or symbols of
    Macaulay2 output labels (looking like 'o39 = ') present on every non-empty
    line.

    Return type: string

    .. NOTE::

        If ``s`` consists of several outputs and their labels have
        different width, it is possible that some strings will have leading
        spaces (or maybe even pieces of output labels). However, this
        function will try not cut any messages.

    EXAMPLES::

        sage: from sage.interfaces.macaulay2 import remove_output_labels
        sage: output = 'o1 = QQ [x, y]\n\no1 : PolynomialRing\n'
        sage: remove_output_labels(output)
        'QQ [x, y]\n\nPolynomialRing\n'
    """
    label = re.compile(r"^o+[0-9]+ (=|:) |^ *")
    lines = s.split("\n")
    it = (label.match(li) for li in lines if li)
    matches = [m for m in it if m is not None]
    if not matches:
        return s
    n = min(m.end() - m.start() for m in matches)
    return "\n".join(li[n:] for li in lines)


PROMPT = "_EGAS_ : "
PROMPT_LOAD = "_EGAS_LOAD_ : "


class Macaulay2(ExtraTabCompletion, Expect):
    """
    Interface to the Macaulay2 interpreter.
    """
    def __init__(self, maxread=None, script_subdirectory=None,
                 logfile=None, server=None,
                 server_tmpdir=None, command=None) -> None:
        """
        Initialize a Macaulay2 interface instance.

        We replace the standard input prompt with a strange one, so that
        we do not catch input prompts inside the documentation.

        We replace the standard input continuation prompt, which is
        just a bunch of spaces and cannot be automatically detected in a
        reliable way. This is necessary for allowing commands that occupy
        several strings.

        We also change the starting line number to make all the output
        labels to be of the same length. This allows correct stripping of
        the output of several commands.

        TESTS::

            sage: from sage.interfaces.macaulay2 import macaulay2
            sage: macaulay2 == loads(dumps(macaulay2))
            True
        """
        if command is None:
            command = os.getenv('SAGE_MACAULAY2_COMMAND') or 'M2'
        init_str = (
            # Prompt changing commands
            'sageLoadMode = false;'
            'ZZ#{Standard,Core#"private dictionary"#"InputPrompt"} = '
            'ZZ#{Standard,Core#"private dictionary"#"InputContinuationPrompt"} = ' +
            'lineno -> if(sageLoadMode) then "%s" else "%s";' % (PROMPT_LOAD, PROMPT) +
            # Also prevent line wrapping in Macaulay2
            "printWidth = 0;" +
            # And make all output labels to be of the same width
            "lineNumber = 10^9;"
            # Assignment of internal expect variables.
            'sageAssign = (k, v) -> (if not instance(v, Sequence) then use v; k <- v);'
        )
        command = "%s --no-debug --no-readline --silent -e '%s'" % (command, init_str)
        Expect.__init__(self,
                        name='macaulay2',
                        prompt=PROMPT,
                        command=command,
                        server=server,
                        server_tmpdir=server_tmpdir,
                        script_subdirectory=script_subdirectory,
                        verbose_start=False,
                        logfile=logfile,
                        eval_using_file_cutoff=500)

    # Macaulay2 provides no "clear" function. However, Macaulay2 does provide
    # garbage collection; since expect automatically reuses variable names,
    # garbage collection in Sage properly sets up garbage collection in
    # Macaulay2.

    def __reduce__(self):
        """
        Used in serializing a Macaulay2 interface.

        EXAMPLES::

            sage: rlm2, t = Macaulay2().__reduce__()
            sage: rlm2(*t)
            Macaulay2
        """
        return reduce_load_macaulay2, tuple()

    def _read_in_file_command(self, filename) -> str:
        """
        Load and *execute* the content of ``filename`` in Macaulay2.

        INPUT:

        - ``filename`` -- string; the name of the file to be loaded and
          executed

        OUTPUT:

        Returns Macaulay2 command loading and executing commands in
        ``filename``.
        Return type: string

        TESTS::

            sage: filename = tmp_filename()
            sage: f = open(filename, "w")
            sage: _ = f.write("sage_test = 7;")
            sage: f.close()
            sage: macaulay2.read(filename)  # indirect doctest
            sage: macaulay2.eval("sage_test")
            7
            sage: import os
            sage: os.unlink(filename)
            sage: macaulay2(10^10000) == 10^10000
            True
        """
        # We use `input` because `load` does not echo the output values
        return 'sageLoadMode=true;input "%s";sageLoadMode=false;' % filename

    def _post_process_from_file(self, s) -> str:
        r"""
        TESTS:

        Check that evaluating using a file gives the same result as without (:issue:`25903`)::

            sage: from sage.interfaces.macaulay2 import remove_output_labels
            sage: s1 = macaulay2._eval_line_using_file('ZZ^2')  # indirect doctest
            sage: s2 = macaulay2._eval_line('ZZ^2', allow_use_file=False)
            sage: remove_output_labels(s1) == remove_output_labels(s2)
            True

        Test multiline input from file::

            sage: (macaulay2.eval('ZZ^2\nZZ^3', allow_use_file=False) ==     # indirect doctest
            ....:     macaulay2.eval('ZZ^2\n%sZZ^3' % (' ' * macaulay2._eval_using_file_cutoff)))
            True
        """
        return '\n'.join(line for line in s.split('\n')
                         if not line.startswith(PROMPT_LOAD))

    def eval(self, code, strip=True, **kwds):
        """
        Send the code x to the Macaulay2 interpreter and return the output
        as a string suitable for input back into Macaulay2, if possible.

        INPUT:

        - ``code`` -- string
        - ``strip`` -- ignored

        EXAMPLES::

            sage: macaulay2.eval("2+2")
            4
        """
        code = code.strip()
        # TODO: in some cases change toExternalString to toString??
        ans = Expect.eval(self, code, strip=strip, **kwds).strip('\n')
        if strip:
            ans = remove_output_labels(ans)
        return AsciiArtString(ans)

    def restart(self) -> None:
        r"""
        Restart Macaulay2 interpreter.

        TESTS::

            sage: macaulay2.restart()
        """
        # If we allow restart to be called as a function, there will be
        # parasitic output
        self.eval("restart")

    def set_seed(self, seed=None):
        r"""
        Set the seed for Macaulay2 interpreter.

        INPUT:

        - ``seed`` -- number (default: ``None``); if ``None``, it
          is set to a random number

        OUTPUT: the new seed

        EXAMPLES::

            sage: m = Macaulay2()
            sage: m.set_seed(123456)
            123456
            sage: [m.random(100) for _ in range(11)]
            [8, 29, 5, 22, 4, 32, 35, 57, 3, 95, 36]
        """
        if seed is None:
            seed = self.rand_seed()
        self.eval('setRandomSeed(%d)' % seed)
        self._seed = seed
        return seed

    class options(GlobalOptions):
        r"""
        Global options for Macaulay2 elements.

        @OPTIONS@

        EXAMPLES::

            sage: macaulay2.options.after_print = True
            sage: A = macaulay2(matrix([[1, 2], [3, 6]])); A
            | 1 2 |
            | 3 6 |
            <BLANKLINE>
                     2       2
            Matrix ZZ  <-- ZZ
            sage: A.kernel()
            image | 2  |
                  | -1 |
            <BLANKLINE>
                                      2
            ZZ-module, submodule of ZZ
            sage: macaulay2.options.after_print = False
        """
        NAME = 'Macaulay2'
        module = 'sage.interfaces.macaulay2'
        after_print = dict(default=False,
                           description='append AfterPrint type information to '
                                       'textual representations',
                           checker=lambda val: isinstance(val, bool))

    def get(self, var):
        """
        Get the value of the variable ``var``.

        INPUT:

        - ``var`` -- string; the name of the variable in Macaulay2

        OUTPUT: string of the textual representation of the variable in
        Macaulay2

        EXAMPLES::

            sage: macaulay2.set("a", "2")
            sage: macaulay2.get("a")
            2

        Note that the following syntax is used to obtain a
        ``Macaulay2Element`` instead::

            sage: a = macaulay2('2'); a
            2
            sage: type(a)
            <class 'sage.interfaces.macaulay2.Macaulay2Element'>
        """
        return self.eval('print(%s)' % var, strip=False)

    def set(self, var, value):
        """
        Set the variable ``var`` to the given value.

        INPUT:

        - ``var`` -- string; the name of the variable in Macaulay2
        - ``value`` -- string to evaluate

        EXAMPLES::

            sage: macaulay2.set("a", "1+1")
            sage: macaulay2.get("a")
            2

        TESTS:

        Check that internal expect variables do not acquire their global
        variable name and that ``use`` is invoked (:issue:`28303`)::

            sage: R = macaulay2('QQ[x, y]')  # indirect doctest
            sage: R.net()
            QQ[x...y]
            sage: S = R / macaulay2('ideal {x^2 - y}')
            sage: macaulay2.eval('class x === %s' % S.name())
            true
        """
        if re.match(r'sage\d+$', var):
            cmd = 'sageAssign(symbol %s,(%s));' % (var, value)
        else:
            cmd = '%s=(%s);' % (var, value)
        ans = Expect.eval(self, cmd, strip=False)
        if ans.find("stdio:") != -1:
            raise RuntimeError("Error evaluating Macaulay2 code.\nIN:%s\nOUT:%s" % (cmd, ans))

    def clear(self, var) -> None:
        """
        Clear the variable named ``var``.

        The interface automatically clears Macaulay2 elements when they fall
        out of use, so calling this method is usually not necessary.

        EXAMPLES::

            sage: macaulay2.eval('R = QQ[x,y];')
            sage: macaulay2.eval('net class R')
            PolynomialRing
            sage: macaulay2.clear('R')
            sage: macaulay2.eval('net class R')
            Symbol

        TESTS:

        Check that only internal variables get reused by the interface::

            sage: all(s.startswith('sage') for s in macaulay2._available_vars)
            True
        """
        if re.match(r'sage\d+$', var):
            self._available_vars.append(var)
        else:
            # this approach is also used by Macaulay2 itself in clearAll
            cmd = 'globalAssign(symbol {0},symbol {0});'.format(var)
            Expect.eval(self, cmd, strip=False)

    def _contains(self, v1, v2) -> bool:
        """
        EXAMPLES::

            sage: a = macaulay2([3,4,5])
            sage: 0 in a, 2 in a, 3 in a
            (True, True, False)
            sage: b = macaulay2('hashTable {"x" => 1, "y" => 2}')
            sage: 'x' in b, '"x"' in b    # indirect doctest
            (False, True)
        """
        return self.eval("%s#?%s" % (v2, v1)) == self._true_symbol()

    def _object_class(self):
        """
        Return the class of Macaulay2 elements.

        EXAMPLES::

            sage: macaulay2._object_class()
            <class 'sage.interfaces.macaulay2.Macaulay2Element'>
        """
        return Macaulay2Element

    def _function_class(self):
        """
        Return the class of Macaulay2 functions.

        EXAMPLES::

            sage: macaulay2._function_class()
            <class 'sage.interfaces.macaulay2.Macaulay2Function'>
        """
        return Macaulay2Function

    def _function_element_class(self):
        """
        Return the class of partially-applied Macaulay2 functions.

        EXAMPLES::

            sage: macaulay2._function_element_class()
            <class 'sage.interfaces.macaulay2.Macaulay2FunctionElement'>
        """
        return Macaulay2FunctionElement

    def console(self):
        """
        Spawn a new M2 command-line session.

        EXAMPLES::

            sage: macaulay2.console()                    # not tested
            Macaulay 2, version 1.1
            with packages: Classic, Core, Elimination, IntegralClosure, LLLBases, Parsing, PrimaryDecomposition, SchurRings, TangentCone
            ...
        """
        macaulay2_console()

    def _install_hints(self) -> str:
        r"""

        TESTS::

            sage: m2 = Macaulay2(command='/wrongpath/M2')
            sage: m2('3+2')
            Traceback (most recent call last):
            ...
            TypeError: unable to start macaulay2 because the command '/wrongpath/M2 ...' failed: The command was not found or was not executable: /wrongpath/M2.
            <BLANKLINE>
                Your attempt to start Macaulay2 failed, either because you do not have
                have Macaulay2 installed, or because it is not configured correctly...
        """
        return r"""
    Your attempt to start Macaulay2 failed, either because you do not have
    have Macaulay2 installed, or because it is not configured correctly.

    - Macaulay2 is not included with Sage, but you can obtain it from
      https://macaulay2.com/.  No additional
      optional Sage packages are required.

    - If you have Macaulay2 installed, then perhaps it is not configured
      correctly. Sage assumes that you can start Macaulay2 with the command
      M2.

    - Alternatively, you can use the following command
      to point Sage to the correct command for your system.

          m2 = Macaulay2(command='/usr/local/bin/M2')

      or by setting the environment variable SAGE_MACAULAY2_COMMAND.
        """

    def _left_list_delim(self) -> str:
        """
        Return the Macaulay2 left delimiter for lists.

        EXAMPLES::

            sage: macaulay2._left_list_delim()
            '{'
        """
        return '{'

    def _right_list_delim(self) -> str:
        """
        Return the Macaulay2 right delimiter for lists.

        EXAMPLES::

            sage: macaulay2._right_list_delim()
            '}'
        """
        return '}'

    def _true_symbol(self) -> str:
        """
        Return the Macaulay2 symbol for True.

        EXAMPLES::

            sage: macaulay2._true_symbol()
            'true'
        """
        return 'true'

    def _false_symbol(self) -> str:
        """
        Return the Macaulay2 symbol for False.

        EXAMPLES::

            sage: macaulay2._false_symbol()
            'false'
        """
        return 'false'

    def _equality_symbol(self) -> str:
        """
        Return the Macaulay2 symbol for equality.

        EXAMPLES::

            sage: macaulay2._false_symbol()
            'false'
        """
        return '=='

    def cputime(self, t=None) -> float:
        """
        EXAMPLES::

            sage: R = macaulay2("QQ[x,y]")
            sage: x,y = R.gens()
            sage: a = (x+y+1)^20
            sage: macaulay2.cputime()       # random
            0.48393700000000001
        """
        _t = float(self.cpuTime()._sage_())
        return _t - t if t else _t

    def version(self) -> tuple[int, ...]:
        """
        Return the version of Macaulay2 as a tuple.

        EXAMPLES::

            sage: macaulay2.version()
            (1, ...)
        """
        s = self.eval('version#"VERSION"')
        return tuple(int(i) for i in s.split("."))

    # Constructors

    def ideal(self, *gens):
        """
        Return the ideal generated by gens.

        INPUT:

        - ``gens`` -- list or tuple of Macaulay2 objects (or objects that can be
          made into Macaulay2 objects via evaluation)

        OUTPUT: the Macaulay2 ideal generated by the given list of gens

        EXAMPLES::

            sage: R2 = macaulay2.ring('QQ', '[x, y]'); R2
            QQ[x...y]
            sage: I = macaulay2.ideal( ('y^2 - x^3', 'x - y') ); I
                      3    2
            ideal (- x  + y , x - y)
            sage: J = I^3; J.gb().gens().transpose()
            {-9} | y9-3y8+3y7-y6             |
            {-7} | xy6-2xy5+xy4-y7+2y6-y5    |
            {-5} | x2y3-x2y2-2xy4+2xy3+y5-y4 |
            {-3} | x3-3x2y+3xy2-y3           |
        """
        if len(gens) == 1 and isinstance(gens[0], (list, tuple)):
            gens = gens[0]
        gens2 = []
        for g in gens:
            if not isinstance(g, Macaulay2Element):
                gens2.append(self(g))
            else:
                gens2.append(g)
        return self('ideal {%s}' % (",".join(g.name() for g in gens2)))

    def ring(self, base_ring='ZZ', vars='[x]', order='Lex'):
        r"""
        Create a Macaulay2 polynomial ring.

        INPUT:

        - ``base_ring`` -- base ring (see examples below)
        - ``vars`` -- tuple or string that defines the variable names
        - ``order`` -- string (default: ``'Lex'``); the monomial order

        OUTPUT: a Macaulay2 ring

        EXAMPLES:

        This is a ring in variables named ``a`` through ``d`` over the finite
        field of order 7, with graded reverse lex ordering::

            sage: R1 = macaulay2.ring('ZZ/7', '[a..d]', 'GRevLex')
            sage: R1.describe()
            ZZ
            --[a..d, Degrees => {4:1}, Heft => {1}, MonomialOrder => {MonomialSize => 16}]
             7                                                       {GRevLex => {4:1}  }
                                                                     {Position => Up    }
            sage: R1.char()
            7

        This is a polynomial ring over the rational numbers::

            sage: R2 = macaulay2.ring('QQ', '[x, y]')
            sage: R2.describe()
            QQ[x..y, Degrees => {2:1}, Heft => {1}, MonomialOrder => {MonomialSize => 16}]
                                                                     {Lex => 2          }
                                                                     {Position => Up    }

        TESTS::

            sage: macaulay2.ring('QQ', '[a_0..a_2,b..<d,f]').vars()
            | a_0 a_1 a_2 b c f |
        """
        return self.new(self._macaulay2_input_ring(base_ring, vars, order))

    def help(self, s):
        """
        EXAMPLES::

            sage: macaulay2.help("load")  # 1st call might be chatty
            ...
            sage: macaulay2.help("load")
            load...
            ****...
            ...
              * "input" -- read Macaulay2 commands and echo
              * "notify" -- whether to notify the user when a file is loaded...

        TESTS:

        Check that help also works for Macaulay2 keywords and variables
        (:issue:`28565`)::

            sage: from sage.repl.interpreter import get_test_shell
            sage: shell = get_test_shell()
            sage: shell.run_cell('macaulay2.help("try")')
            try -- catch an error
            ****...
            The object "try" is a...

            sage: from sage.repl.interpreter import get_test_shell
            sage: shell = get_test_shell()
            sage: shell.run_cell('macaulay2.help("errorDepth")')
            errorDepth...
            The object "errorDepth" is a...
        """
        r = self.eval('help "%s"' % s)
        end = r.rfind("\n\nDIV")
        if end != -1:
            r = r[:end]
        return AsciiArtString(r)

    def _tab_completion(self) -> list[str]:
        """
        Return a list of tab completions for Macaulay2.

        Returns dynamically built sorted list of commands obtained using
        Macaulay2 "apropos" command.
        Return type: list of strings

        TESTS::

            sage: names = macaulay2._tab_completion()
            sage: 'ring' in names
            True
            sage: macaulay2.eval("abcabc = 4")
            4
            sage: names = macaulay2._tab_completion()
            sage: "abcabc" in names
            True
        """
        # Get all the names from Macaulay2 except numbered outputs like
        # o1, o2, etc. and automatic Sage variable names sage0, sage1, etc.
        # It is faster to get it back as a string.
        r = macaulay2.eval(r"""
            print toString select(
                apply(apropos "^[[:alnum:]]+$", toString),
                s -> not match("^(o|sage)[0-9]+$", s))
            """)
        # Now split this string into separate names
        # Macaulay2 sorts things like A, a, B, b, ...
        return sorted(r[1:-1].split(", "))

    def use(self, R) -> None:
        """
        Use the Macaulay2 ring R.

        EXAMPLES::

            sage: R = macaulay2("QQ[x,y]")
            sage: P = macaulay2("ZZ/7[symbol x, symbol y]")
            sage: macaulay2("x").cls()._operator('===', P)
            true
            sage: macaulay2.use(R)
            sage: macaulay2("x").cls()._operator('===', R)
            true
        """
        R = self(R)
        self.eval("use %s;" % R.name(), strip=False)

    def new_from(self, type, value):
        """
        Return a new ``Macaulay2Element`` of type ``type`` constructed from
        ``value``.

        EXAMPLES::

            sage: l = macaulay2.new_from("MutableList", [1,2,3])
            sage: l
            MutableList{...3...}
            sage: list(l)
            [1, 2, 3]
        """
        type = self(type)
        value = self(value)
        return self.new("new %s from %s" % (type.name(), value.name()))

    def _macaulay2_input_ring(self, base_ring, vars, order='GRevLex'):
        """
        Build a string representation of a polynomial ring which can be used as
        Macaulay2 input.

        TESTS::

            sage: R = GF(101)['x']
            sage: macaulay2._macaulay2_input_ring(R.base_ring(), R.gens(), 'Lex')
            'sage...[symbol x, MonomialSize=>16, MonomialOrder=>Lex]'
        """
        if not isinstance(base_ring, str):
            base_ring = self(base_ring).name()

        varstr = str(vars)[1:-1].rstrip(',')
        r = re.compile(r"(?<=,)|(?<=\.\.<)|(?<=\.\.)(?!<)")
        varstr = "symbol " + r.sub("symbol ", varstr)
        return '%s[%s, MonomialSize=>16, MonomialOrder=>%s]' % (base_ring, varstr,
                                                                order)


@instancedoc
class Macaulay2Element(ExtraTabCompletion, ExpectElement, sage.interfaces.abc.Macaulay2Element):
    """
    Instances of this class represent objects in Macaulay2.

    Using the method :meth:`sage` we can translate some of them to
    SageMath objects:

    .. automethod:: _sage_
    """
    def _latex_(self) -> str:
        r"""
        EXAMPLES::

            sage: m = macaulay2('matrix {{1,2},{3,4}}')
            sage: m
            | 1 2 |
            | 3 4 |
            sage: latex(m)
            \left(\!\begin{array}{cc}\n1&2\\\n3&4\n\end{array}\!\right)
        """
        s = self.tex().external_string().strip('"').strip('$').replace('\\\\', '\\')
        s = s.replace(r"\bgroup", "").replace(r"\egroup", "")
        return s

    def __iter__(self) -> Iterator:
        """
        EXAMPLES::

            sage: l = macaulay2([1,2,3])
            sage: list(iter(l))
            [1, 2, 3]
        """
        for i in range(len(self)):  # zero-indexed!
            yield self[i]

    def __str__(self) -> str:
        """
        EXAMPLES::

            sage: R = macaulay2("QQ[x,y,z]/(x^3-y^3-z^3)")
            sage: x = macaulay2('x')
            sage: y = macaulay2('y')
            sage: str(x+y)
            x + y
            sage: str(macaulay2("QQ[x,y,z]"))
            QQ[x...z]
            sage: str(macaulay2("QQ[x,y,z]/(x+y+z)"))
             QQ[x...z]
            -------...
            x + y + z
        """
        P = self._check_valid()
        return P.get(self._name)

    def _repr_(self) -> str:
        """
        EXAMPLES::

            sage: repr(macaulay2('1..25'))
            (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22,
            --------------------------------------------------------------------------------
            23, 24, 25)
            sage: str(macaulay2('1..25'))
            (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25)

        If ``AfterPrint`` is enabled, the ``repr`` contains type information,
        but the string representation does not::

            sage: macaulay2.options.after_print = True
            sage: repr(macaulay2('1..25'))
            (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22,
            --------------------------------------------------------------------------------
            23, 24, 25)
            <BLANKLINE>
            Sequence
            sage: str(macaulay2('1..25'))
            (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25)
            sage: macaulay2.options.after_print = False
        """
        from sage.typeset.ascii_art import empty_ascii_art
        P = self.parent()
        if P.options.after_print:
            # In M2, the wrapped output is indented by the width of the prompt,
            # which we strip in Sage. We hardcode the width of the prompt to
            # 14=len('o1000000001 = '), which is tested in the doctests by the
            # output getting wrapped at 80 characters.
            width = 14 + empty_ascii_art._terminal_width()
            return P.eval('printWidth=%d;%s' % (width, self._name))
        # Otherwise manually wrap the net representation which does not display
        # AfterPrint text
        return P.eval('print(wrap(%d,"-",net %s))'
                      % (empty_ascii_art._terminal_width(), self._name),
                      strip=False)

    def external_string(self) -> str:
        """
        EXAMPLES::

           sage: R = macaulay2("QQ[symbol x, symbol y]")
           sage: R.external_string()
           'QQ[x..y, Degrees => {2:1}, Heft => {1}]'
        """
        P = self._check_valid()
        code = 'toExternalString(%s)' % self.name()
        X = P.eval(code, strip=True)

        if 'stdio:' in X:
            if 'to external string' in X:
                return P.eval('%s' % self.name())
            raise RuntimeError("Error evaluating Macaulay2 code.\nIN:%s\nOUT:%s" % (code, X))

        return multiple_replace({'\r': '', '\n': ' '}, X)

    def name(self, new_name=None):
        """
        Get or change the name of this Macaulay2 element.

        INPUT:

        - ``new_name`` -- string (default: ``None``); if ``None``, return the
          name of this element. Else return a new object identical to ``self``
          whose name is ``new_name``.

        Note that this can overwrite existing variables in the system.

        EXAMPLES::

            sage: S = macaulay2(QQ['x,y'])
            sage: S.name()
            'sage...'
            sage: R = S.name("R")
            sage: R.name()
            'R'
            sage: R.vars().cokernel().resolution()
             1      2      1
            R  <-- R  <-- R  <-- 0
            <BLANKLINE>
            0      1      2      3

        The name can also be given at definition::

            sage: A = macaulay2(ZZ['x,y,z'], name='A')
            sage: A.name()
            'A'
            sage: A^1
             1
            A
        """
        if new_name is None:
            return self._name
        if not isinstance(new_name, str):
            raise TypeError("new_name must be a string")

        P = self.parent()
        # First release self, so that new_name becomes the initial reference to
        # its value.  This is needed to change the name of a PolynomialRing.
        # NOTE: This does not work if self._name is not the initial reference.
        cmd = """(() -> (
            m := lookup(GlobalReleaseHook, class {0});
            if m =!= null then m(symbol {0}, {0});
            {1} = {0};
            ))()""".format(self._name, new_name)
        ans = P.eval(cmd)
        if ans.find("stdio:") != -1:
            raise RuntimeError("Error evaluating Macaulay2 code.\n"
                               "IN:%s\nOUT:%s" % (cmd, ans))
        return P._object_class()(P, new_name, is_name=True)

    def __len__(self) -> int:
        """
        EXAMPLES::

            sage: l = macaulay2([1,2,3])
            sage: len(l)
            3
            sage: type(_)
            <... 'int'>
        """
        self._check_valid()
        # we use str instead of repr to avoid wrapping
        return int(str(self.parent()("#%s" % self.name())))

    def __getitem__(self, n):
        """
        EXAMPLES::

            sage: l = macaulay2([1,2,3])
            sage: l[0]
            1
        """
        self._check_valid()
        n = self.parent()(n)
        return self.parent().new('%s # %s' % (self.name(), n.name()))

    def __setitem__(self, index, value):
        """
        EXAMPLES::

            sage: l = macaulay2.new_from("MutableList", [1,2,3])
            sage: l[0] = 4
            sage: list(l)
            [4, 2, 3]
        """
        P = self.parent()
        index = P(index)
        value = P(value)
        res = P.eval("%s # %s = %s" % (self.name(), index.name(), value.name()))
        if "assignment attempted to element of immutable list" in res:
            raise TypeError("item assignment not supported")

    def __call__(self, x):
        """
        EXAMPLES::

            sage: R = macaulay2("QQ[x, y]")
            sage: x,y = R.gens()
            sage: I = macaulay2.ideal(x*y, x+y)
            sage: gb = macaulay2.gb
            sage: gb(I)
            GroebnerBasis[status: done; S-pairs encountered up to degree 1]
        """
        self._check_valid()
        P = self.parent()
        r = P(x)
        return P('%s %s' % (self.name(), r.name()))

    def __floordiv__(self, x):
        """
        Quotient of division of ``self`` by ``other``.  This is denoted ``//``.

        EXAMPLES::

            sage: R.<x,y> = GF(7)[]

        Now make the M2 version of R, so we can coerce elements of R to M2::

            sage: _ = macaulay2(R)
            sage: h = macaulay2((x^3 + 2*y^2*x)^7); h
             21     7 14
            x   + 2x y
            sage: h1 = macaulay2(x^2 + 2*y*x)
            sage: h2 = macaulay2(x^3 + 2*y*x)
            sage: u = h // [h1,h2]
            sage: h == u[0]*h1 + u[1]*h2 + (h % [h1,h2])
            True
        """
        if isinstance(x, (list, tuple)):
            y = self.parent(x)
            z = self.parent().new('%s // matrix{%s}' % (self.name(), y.name()))
            return list(z.entries().flatten())
        return self.parent().new('%s // %s' % (self.name(), x.name()))

    def __mod__(self, x):
        """
        Remainder of division of ``self`` by ``other``.  This is denoted ``%``.

        EXAMPLES::

            sage: R.<x,y> = GF(7)[]

        Now make the M2 version of R, so we can coerce elements of R to M2::

            sage: _ = macaulay2(R)
            sage: h = macaulay2((x^3 + 2*y^2*x)^7); h
             21     7 14
            x   + 2x y
            sage: h1 = macaulay2(x^2 + 2*y*x)
            sage: h2 = macaulay2(x^3 + 2*y*x)
            sage: h % [h1,h2]
            -3x*y
            sage: u = h // [h1,h2]
            sage: h == u[0]*h1 + u[1]*h2 + (h % [h1,h2])
            True
        """
        if isinstance(x, (list, tuple)):
            y = self.parent(x)
            return self.parent().new('%s %% matrix{%s}' % (self.name(), y.name()))
        if not isinstance(x, Macaulay2Element):
            x = self.parent(x)
        return self.parent().new('%s %% %s' % (self.name(), x.name()))

    def __bool__(self) -> bool:
        """
        Return whether this Macaulay2 element is not ``False`` or not ``0``.

        EXAMPLES::

            sage: a = macaulay2(0)
            sage: a == 0
            True
            sage: bool(a)
            False

        TESTS:

        Check that :issue:`28705` is fixed::

            sage: t = macaulay2(True); t
            true
            sage: bool(t)
            True
            sage: bool(macaulay2('false'))
            False
            sage: bool(macaulay2('"a"'))
            True
        """
        P = self.parent()
        return P.eval('{0}===false or {0}==0'.format(self._name)) != 'true'

    def sage_polystring(self) -> str:
        """
        If this Macaulay2 element is a polynomial, return a string
        representation of this polynomial that is suitable for
        evaluation in Python.  Thus ``*`` is used for multiplication
        and ``**`` for exponentiation.   This function is primarily
        used internally.

        EXAMPLES::

            sage: R = macaulay2.ring('QQ','(x,y)')
            sage: f = macaulay2('x^3 + 3*y^11 + 5')
            sage: print(f)
             3     11
            x  + 3y   + 5
            sage: f.sage_polystring()
            'x**3+3*y**11+5'
        """
        return self.external_string().replace('^', '**')

    def subs(self, *args, **kwds):
        """
        Note that we have to override the substitute method so that we get
        the default one from Macaulay2 instead of the one provided by Element.

        EXAMPLES::

            sage: R = macaulay2("QQ[x]")
            sage: P = macaulay2("ZZ/7[symbol x]")
            sage: x, = R.gens()
            sage: a = x^2 + 1
            sage: a = a.substitute(P)
            sage: a.sage().parent()
            Univariate Polynomial Ring in x over Finite Field of size 7
        """
        return self.__getattr__("substitute")(*args, **kwds)

    def _tab_completion(self) -> list[str]:
        """
        Return a list of tab completions for ``self``.

        Returns dynamically built sorted list of commands obtained using
        Macaulay2 "methods" command. All returned functions can take ``self``
        as their first argument

        Return type: list of strings

        TESTS::

            sage: a = macaulay2("QQ[x,y]")
            sage: traits = a._tab_completion()
            sage: "generators" in traits
            True

        The implementation of this function does not set or change global
        variables::

            sage: a.dictionary()._operator('#?', '"r"')
            false
        """
        # It is possible, that these are not all possible methods, but
        # there are still plenty and at least there are no definitely
        # wrong ones...
        r = self.parent().eval(
            """(() -> (
            currentClass := class %s;
            total := {};
            while true do (
                -- Select methods with first argument of the given class
                r := select(methods currentClass, s -> s_1 === currentClass);
                -- Get their names as strings
                r = apply(r, s -> toString s_0);
                -- Keep only alpha-numeric ones
                r = select(r, s -> match("^[[:alnum:]]+$", s));
                -- Add to existing ones
                total = total | select(r, s -> not any(total, e -> e == s));
                if parent currentClass === currentClass then break;
                currentClass = parent currentClass;
                );
            print toString total
            ))()""" % self.name())
        r = sorted(r[1:-1].split(", "))
        return r

    def cls(self):
        """
        Since class is a keyword in Python, we have to use cls to call
        Macaulay2's class.  In Macaulay2, class corresponds to Sage's
        notion of parent.

        EXAMPLES::

            sage: macaulay2(ZZ).cls()
            Ring
        """
        return self.parent()(f"class {self.name()}")

    def after_print_text(self):
        r"""
        Obtain type information for this Macaulay2 element.

        This is the text that is displayed using ``AfterPrint`` in a Macaulay2
        interpreter.

        Macaulay2 by default includes this information in the output.
        In Sage, this behavior can optionally be enabled by setting the option
        ``after_print`` in :class:`Macaulay2.options`.

        EXAMPLES::

            sage: B = macaulay2(matrix([[1, 2], [3, 6]])).kernel(); B
            image | 2  |
                  | -1 |
            sage: B.after_print_text()
                                      2
            ZZ-module, submodule of ZZ
        """
        return self.parent().eval('(lookup({topLevelMode,AfterPrint},' +
                                  'class {0}))({0})'.format(self._name))

    ############################
    # Aliases for M2 operators #
    ############################
    def dot(self, x):
        """
        EXAMPLES::

            sage: d = macaulay2.new("MutableHashTable")
            sage: d["k"] = 4
            sage: d.dot("k")
            4
        """
        parent = self.parent()
        x = parent(x)
        return parent("%s.%s" % (self.name(), x))

    def _operator(self, opstr: str, x):
        """
        Return the infix binary operation specified by opstr applied
        to ``self`` and x.

        EXAMPLES::

            sage: a = macaulay2("3")
            sage: a._operator("+", a)
            6
            sage: a._operator("*", a)
            9
        """
        parent = self.parent()
        x = parent(x)
        return parent("%s%s%s" % (self.name(), opstr, x.name()))

    def sharp(self, x):
        """
        EXAMPLES::

            sage: a = macaulay2([1,2,3])
            sage: a.sharp(0)
            1
        """
        return self._operator("#", x)

    def starstar(self, x):
        """
        The binary operator ``**`` in Macaulay2 is usually used for tensor
        or Cartesian power.

        EXAMPLES::

            sage: a = macaulay2([1,2]).set()
            sage: a.starstar(a)
            set {(1, 1), (1, 2), (2, 1), (2, 2)}
        """
        return self._operator("**", x)

    def underscore(self, x):
        """
        EXAMPLES::

            sage: a = macaulay2([1,2,3])
            sage: a.underscore(0)
            1
        """
        return self._operator("_", x)

    ######################
    # Conversion to Sage #
    ######################
    def _sage_(self):
        r"""
        EXAMPLES::

            sage: macaulay2(ZZ).sage()         # indirect doctest
            Integer Ring
            sage: macaulay2(QQ).sage()
            Rational Field

            sage: macaulay2(2).sage()
            2
            sage: macaulay2(1/2).sage()
            1/2
            sage: macaulay2(2/1).sage()
            2
            sage: _.parent()
            Rational Field
            sage: macaulay2([1,2,3]).sage()
            [1, 2, 3]

            sage: m = matrix([[1,2],[3,4]])
            sage: macaulay2(m).sage()
            [1 2]
            [3 4]

            sage: D = macaulay2('hashTable {4 => 1, 2 => 3}')
            sage: D.pairs()
            {(4, 1), (2, 3)}
            sage: D.sage() == {4: 1, 2: 3}
            True

            sage: macaulay2(QQ['x,y']).sage()
            Multivariate Polynomial Ring in x, y over Rational Field
            sage: macaulay2(QQ['x']).sage()
            Univariate Polynomial Ring in x over Rational Field
            sage: macaulay2(GF(7)['x,y']).sage()
            Multivariate Polynomial Ring in x, y over Finite Field of size 7

            sage: macaulay2(GF(7)).sage()
            Finite Field of size 7
            sage: macaulay2(GF(49, 'a')).sage()
            Finite Field in a of size 7^2

            sage: R.<x,y> = QQ[]
            sage: macaulay2(x^2+y^2+1).sage()
            x^2 + y^2 + 1

            sage: R = macaulay2("QQ[x,y]")
            sage: I = macaulay2("ideal (x,y)")
            sage: I.sage()
            Ideal (x, y) of Multivariate Polynomial Ring in x, y over Rational Field

            sage: macaulay2("x = symbol x")
            x
            sage: macaulay2("QQ[x_0..x_25]").sage()
            Multivariate Polynomial Ring in x_0, x_1,..., x_25 over Rational Field

            sage: S = ZZ['x,y'].quotient('x^2-y')
            sage: macaulay2(S).sage() == S
            True
            sage: S = GF(101)['x,y'].quotient('x^2-y')
            sage: macaulay2(S).sage() == S
            True

            sage: R = GF(13)['a,b']['c,d']
            sage: macaulay2(R).sage() == R
            True
            sage: macaulay2('a^2 + c').sage() == R('a^2 + c')
            True
            sage: macaulay2.substitute('a', R).sage().parent() is R
            True

            sage: R = macaulay2("QQ^2")
            sage: R.sage()
            Vector space of dimension 2 over Rational Field

            sage: macaulay2("vector {4_QQ, 2}").sage()
            (4, 2)
            sage: _.parent()
            Vector space of dimension 2 over Rational Field

            sage: m = macaulay2('"hello"')
            sage: m.sage()
            'hello'

            sage: gg = macaulay2.needsPackage('"Graphs"')
            sage: g = macaulay2.barbellGraph(3)
            sage: g.sage()
            Graph on 6 vertices
            sage: g.sage().edges(labels=False)
            [(0, 1), (0, 2), (1, 2), (2, 3), (3, 4), (3, 5), (4, 5)]

            sage: d = 'digraph ({{1,2},{2,1},{3,1}}, EntryMode => "edges")'
            sage: g = macaulay2(d)
            sage: g.sage()
            Digraph on 3 vertices
            sage: g.sage().edges(labels=False)
            [(1, 2), (2, 1), (3, 1)]

        Chain complexes and maps of chain complexes can be converted::

            sage: R = ZZ['a,b,c']
            sage: C = macaulay2(ideal(R.gens())).resolution()
            sage: unicode_art(C.sage())
                                  ⎛-b  0 -c⎞     ⎛ c⎞
                                  ⎜ a -c  0⎟     ⎜ a⎟
                      (a b c)     ⎝ 0  b  a⎠     ⎝-b⎠
            0 <── C_0 <────── C_1 <───────── C_2 <─── C_3 <── 0
            sage: F = C.dot('dd')
            sage: G = F.sage()
            sage: G.in_degree(2)
            [-b  0 -c]
            [ a -c  0]
            [ 0  b  a]
            sage: F.underscore(2).sage() == G.in_degree(2)
            True
            sage: (F^2).sage()
            Chain complex morphism:
              From: Chain complex with at most 4 nonzero terms over Multivariate Polynomial Ring in a, b, c over Integer Ring
              To:   Chain complex with at most 4 nonzero terms over Multivariate Polynomial Ring in a, b, c over Integer Ring

        Quotient rings in Macaulay2 inherit variable names from the ambient
        ring, so we mimic this behaviour in Sage::

            sage: R = macaulay2("ZZ/7[x,y]")
            sage: I = macaulay2("ideal (x^3 - y^2)")
            sage: (R/I).gens()
            {x, y}
            sage: (R/I).sage().gens()
            (x, y)

        Elements of quotient rings::

            sage: x, y = (R/I).gens()
            sage: f = ((x^3 + 2*y^2*x)^7).sage(); f
            2*x*y^18 + y^14
            sage: f.parent()
            Quotient of Multivariate Polynomial Ring in x, y over Finite Field of size 7 by the ideal (x^3 - y^2)
        """
        repr_str = str(self)
        cls_str = str(self.cls())
        cls_cls_str = str(self.cls().cls())

        if repr_str == "ZZ":
            from sage.rings.integer_ring import ZZ
            return ZZ
        if repr_str == "QQ":
            from sage.rings.rational_field import QQ
            return QQ

        if cls_cls_str == "Type":
            if cls_str == "List":
                return [entry._sage_() for entry in self]
            if cls_str == "Matrix":
                base_ring = self.ring()._sage_()
                return self._matrix_(base_ring)
            if cls_str == 'HashTable':
                return {x._sage_(): y._sage_() for x, y in self.pairs()}
            if cls_str == "Ideal":
                parent = self.ring()._sage_()
                gens = self.gens().entries().flatten()._sage_()
                return parent.ideal(*gens)
            if cls_str == "QuotientRing":
                # Handle the ZZ/n case
                ambient = self.ambient()
                if ambient.external_string() == 'ZZ':
                    from sage.rings.finite_rings.finite_field_constructor import GF
                    from sage.rings.integer_ring import ZZ
                    external_string = self.external_string()
                    zz, n = external_string.split("/")

                    # Note that n must be prime since it is
                    # coming from Macaulay 2
                    return GF(ZZ(n))

                ambient_ring = ambient._sage_()
                ideal = self.ideal()._sage_()
                return ambient_ring.quotient(ideal, names=ambient_ring.variable_names())
            if cls_str == "PolynomialRing":
                from sage.rings.polynomial.polynomial_ring_constructor import (
                    PolynomialRing,
                )
                from sage.rings.polynomial.term_order import inv_macaulay2_name_mapping

                # Get the base ring
                base_ring = self.coefficientRing()._sage_()

                # Get a string list of generators
                gens = str(self.gens().toString())[1:-1]

                # Check that we are dealing with default degrees, i.e. 1's.
                if self.options().sharp("Degrees").any("x -> x != {1}")._sage_():
                    raise ValueError("cannot convert Macaulay2 polynomial ring with non-default degrees to Sage")
                # Handle the term order
                external_string = self.external_string()
                order = None
                if "MonomialOrder" not in external_string:
                    order = "degrevlex"
                else:
                    for order_name in inv_macaulay2_name_mapping:
                        if order_name in external_string:
                            order = inv_macaulay2_name_mapping[order_name]
                if len(gens) > 1 and order is None:
                    raise ValueError("cannot convert Macaulay2's term order to a Sage term order")

                return PolynomialRing(base_ring, order=order, names=gens)
            if cls_str == "GaloisField":
                from sage.rings.finite_rings.finite_field_constructor import GF
                from sage.rings.integer_ring import ZZ
                gf, n = repr_str.split(" ")
                n = ZZ(n)
                if n.is_prime():
                    return GF(n)
                gen = str(self.gens())[1:-1]
                return GF(n, gen)
            if cls_str == "Boolean":
                if repr_str == "true":
                    return True
                if repr_str == "false":
                    return False
            if cls_str == "String":
                return str(repr_str)
            if cls_str == "Module":
                from sage.modules.free_module import FreeModule
                if self.isFreeModule()._sage_():
                    ring = self.ring()._sage_()
                    rank = self.rank()._sage_()
                    return FreeModule(ring, rank)
            if cls_str in ("Graph", "Digraph"):
                if cls_str == "Graph":
                    from sage.graphs.graph import Graph
                    graph_cls = Graph
                else:
                    from sage.graphs.digraph import DiGraph
                    graph_cls = DiGraph
                adj_mat = self.adjacencyMatrix().sage()
                g = graph_cls(adj_mat, format='adjacency_matrix')
                g.relabel(self.vertices())
                return g
            if cls_str == "ChainComplex":
                from sage.homology.chain_complex import ChainComplex
                ring = self.ring()._sage_()
                dd = self.dot('dd')
                degree = dd.degree()._sage_()
                a = self.min()._sage_()
                b = self.max()._sage_()
                matrices = {i: dd.underscore(i)._matrix_(ring)
                            for i in range(a, b + 1)}
                return ChainComplex(matrices, degree=degree)
            if cls_str == "ChainComplexMap":
                from sage.homology.chain_complex_morphism import ChainComplexMorphism
                ring = self.ring()._sage_()
                source = self.source()
                a = source.min()._sage_()
                b = source.max()._sage_()
                degree = self.degree()._sage_()
                matrices = {i: self.underscore(i)._matrix_(ring)
                            for i in range(a, b + 1)}
                C = source._sage_()
                # in Sage, chain complex morphisms are degree-preserving,
                # so we shift the degrees of the target
                D = self.target()._operator(' ', '[%s]' % degree)._sage_()
                return ChainComplexMorphism(matrices, C, D)
        else:
            # Handle the integers and rationals separately
            if cls_str == "ZZ":
                from sage.rings.integer_ring import ZZ
                return ZZ(repr_str)
            if cls_str == "QQ":
                from sage.rings.rational_field import QQ
                return QQ((self.numerator()._sage_(),
                           self.denominator()._sage_()))

            m2_parent = self.cls()
            parent = m2_parent._sage_()

            if cls_cls_str == "PolynomialRing":
                # going through a dict
                if len(m2_parent.gens()) == 1:
                    d = {monome[0].sage(): coeff.sage()
                         for monome, coeff in self.listForm()}
                else:
                    d = {tuple(monome.sage()): coeff.sage()
                         for monome, coeff in self.listForm()}
                return parent(d)
            if cls_cls_str == "QuotientRing":
                return parent(self.external_string())
            if cls_cls_str == "Module":
                entries = self.entries()._sage_()
                return parent._element_constructor_(entries)

        from sage.misc.sage_eval import sage_eval
        try:
            return sage_eval(repr_str)
        except Exception:
            raise NotImplementedError(f"cannot convert\n{repr_str}\nto a Sage object")

    def _matrix_(self, R):
        r"""
        If ``self`` is a Macaulay2 matrix, return the corresponding Sage matrix
        over the Sage ring ``R``.

        INPUT:

        - ``R`` -- ring to coerce into

        OUTPUT: matrix

        EXAMPLES::

            sage: A = macaulay2('matrix {{1,2},{3,4}}')
            sage: matrix(QQ, A)                     # indirect doctest
            [1 2]
            [3 4]

        TESTS:

        Check that degenerate matrix dimensions are preserved (:issue:`28591`)::

            sage: m = macaulay2('matrix {{},{}}')
            sage: matrix(ZZ, m).dimensions()
            (2, 0)
            sage: matrix(ZZ, m.transpose()).dimensions()
            (0, 2)
        """
        from sage.matrix.constructor import matrix
        m = matrix(R, self.entries()._sage_())
        if not m.nrows():
            return matrix(R, 0, self.numcols()._sage_())
        return m


@instancedoc
class Macaulay2Function(ExpectFunction):
    """
    TESTS::

        sage: gb = macaulay2.gb
        sage: type(gb)
        <class 'sage.interfaces.macaulay2.Macaulay2Function'>
        sage: gb._name
        'gb'
    """

    def _instancedoc_(self):
        """
        EXAMPLES::

            sage: print(macaulay2.load.__doc__)
            nodetex,noreplace
            load...
            ****...
            ...
              * "input" -- read Macaulay2 commands and echo
              * "notify" -- whether to notify the user when a file is loaded...

        TESTS:

        Check that detex is disabled, so that the output does not get
        reformatted (:issue:`28565`)::

            sage: from sage.repl.interpreter import get_test_shell
            sage: shell = get_test_shell()
            sage: shell.run_cell('macaulay2.matrix?')
            ...
            +--------------------------------+
            |  i1 : matrix{{1,2,3},{4,5,6}}  |
            |                                |
            |  o1 = | 1 2 3 |                |
            |       | 4 5 6 |                |
            |                                |
            |                2        3      |
            |  o1 : Matrix ZZ  <--  ZZ       |
            +--------------------------------+
            ...
        """
        r = self._parent.help(self._name)
        return AsciiArtString('nodetex,noreplace\n' + r)

    def _sage_src_(self):
        """
        EXAMPLES::

            sage: macaulay2.gb._sage_src_()
            -- code for method: gb(Ideal)...
            -- code for method: gb(Matrix)...
            ...
        """
        return self._parent.eval('code methods %s' % self._name)


@instancedoc
class Macaulay2FunctionElement(FunctionElement):
    def _instancedoc_(self):
        """
        TESTS:

        Since :issue:`28565`, the help output includes all documentation nodes
        that can take ``self._obj`` as first argument. This also checks that
        detex is disabled, so that the output does not get reformatted. ::

            sage: from sage.repl.interpreter import get_test_shell
            sage: shell = get_test_shell()
            sage: shell.run_cell('I = macaulay2("ideal {4}")')
            sage: shell.run_cell('I.resolution?')
            Signature:       I.resolution(Ideal)
            Call signature:  I.resolution(self, *args, **kwds)
            Type:            Macaulay2FunctionElement
            String form:     resolution
            File:        ...
            Docstring:
            resolution -- projective resolution
            ****...
            <BLANKLINE>
            resolution(Ideal) -- compute a projective resolution of...
            ****...
            |        1      4      6      4      1        |
            |  o3 = R  <-- R  <-- R  <-- R  <-- R  <-- 0  |
            |                                             |
            |       0      1      2      3      4      5  |
            ...
        """
        P = self._obj.parent()
        r = P.eval('help prepend({0}, select(methods {0}, m->'
                   'instance({1}, m#1)))'.format(self._name, self._obj._name))
        end = r.rfind("\n\nDIV")
        if end != -1:
            r = r[:end]
        return AsciiArtString('nodetex,noreplace\n' + r)

    def _sage_src_(self):
        """
        EXAMPLES::

            sage: m = macaulay2('matrix {{4,6}}')
            sage: m.resolution._sage_src_()
            -- code for method: freeResolution(Matrix)...
        """
        return self._obj.parent().eval(
            'code select(methods %s, m->instance(%s, m#1))'
            % (self._name, self._obj._name))


# An instance
macaulay2 = Macaulay2()


def macaulay2_console():
    """
    Spawn a new M2 command-line session.

    EXAMPLES::

        sage: macaulay2_console()                    # not tested
        Macaulay 2, version 1.1
        with packages: Classic, Core, Elimination, IntegralClosure, LLLBases, Parsing, PrimaryDecomposition, SchurRings, TangentCone
        ...
    """
    from sage.repl.rich_output.display_manager import get_display_manager
    if not get_display_manager().is_in_terminal():
        raise RuntimeError('Can use the console only in the terminal. Try %%macaulay2 magics instead.')
    os.system('M2')


def reduce_load_macaulay2():
    """
    Used for reconstructing a copy of the Macaulay2 interpreter from a pickle.

    EXAMPLES::

        sage: from sage.interfaces.macaulay2 import reduce_load_macaulay2
        sage: reduce_load_macaulay2()
        Macaulay2
    """
    return macaulay2



# sage.doctest: needs sage.symbolic
#
# distutils: sources = sage/stats/distributions/dgs_gauss_mp.c sage/stats/distributions/dgs_gauss_dp.c sage/stats/distributions/dgs_bern.c
# distutils: depends = sage/stats/distributions/dgs_gauss.h sage/stats/distributions/dgs_bern.h sage/stats/distributions/dgs_misc.h
# distutils: extra_compile_args = -D_XOPEN_SOURCE=600

r"""
Discrete Gaussian Samplers over the Integers

This class realizes oracles which returns integers proportionally to
`\exp(-(x-c)^2/(2σ^2))`. All oracles are implemented using rejection sampling.
See :func:`DiscreteGaussianDistributionIntegerSampler.__init__` for which algorithms are
available.

AUTHORS:

- Martin Albrecht (2014-06-28): initial version

EXAMPLES:

We construct a sampler for the distribution `D_{3,c}` with width `σ=3` and center `c=0`::

    sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
    sage: sigma = 3.0
    sage: D = DiscreteGaussianDistributionIntegerSampler(sigma=sigma)

We ask for 100000 samples::

    sage: from collections import defaultdict
    sage: counter = defaultdict(Integer)
    sage: n = 0
    sage: def add_samples(i):
    ....:     global counter, n
    ....:     for _ in range(i):
    ....:         counter[D()] += 1
    ....:         n += 1

    sage: add_samples(100000)

These are sampled with a probability proportional to `\exp(-x^2/18)`. More
precisely we have to normalise by dividing by the overall probability over all
integers. We use the fact that hitting anything more than 6 standard deviations
away is very unlikely and compute::

    sage: bound = (6*sigma).floor()
    sage: norm_factor = sum([exp(-x^2/(2*sigma^2)) for x in range(-bound,bound+1)])
    sage: norm_factor
    7.519...

With this normalisation factor, we can now test if our samples follow the
expected distribution::

    sage: expected = lambda x : ZZ(round(n*exp(-x^2/(2*sigma^2))/norm_factor))
    sage: observed = lambda x : counter[x]

    sage: add_samples(10000)
    sage: while abs(observed(0)*1.0/expected(0)   - 1.0) > 5e-2: add_samples(10000)
    sage: while abs(observed(4)*1.0/expected(4)   - 1.0) > 5e-2: add_samples(10000)
    sage: while abs(observed(-10)*1.0/expected(-10) - 1.0) > 5e-2: add_samples(10000)  # long time

We construct an instance with a larger width::

    sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
    sage: sigma = 127
    sage: D = DiscreteGaussianDistributionIntegerSampler(sigma=sigma, algorithm='uniform+online')

ask for 100000 samples::

    sage: from collections import defaultdict
    sage: counter = defaultdict(Integer)
    sage: n = 0
    sage: def add_samples(i):
    ....:     global counter, n
    ....:     for _ in range(i):
    ....:         counter[D()] += 1
    ....:         n += 1

    sage: add_samples(100000)

and check if the proportions fit::

    sage: expected = lambda x, y: (
    ....:     exp(-x^2/(2*sigma^2))/exp(-y^2/(2*sigma^2)).n())
    sage: observed = lambda x, y: float(counter[x])/counter[y]

    sage: while not all(v in counter for v in (0, 1, -100)): add_samples(10000)

    sage: while abs(expected(0, 1) - observed(0, 1)) > 2e-1: add_samples(10000)
    sage: while abs(expected(0, -100) - observed(0, -100)) > 2e-1: add_samples(10000)

We construct a sampler with `c\%1 != 0`::

    sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
    sage: sigma = 3
    sage: D = DiscreteGaussianDistributionIntegerSampler(sigma=sigma, c=1/2)
    sage: s = 0
    sage: n = 0
    sage: def add_samples(i):
    ....:     global s, n
    ....:     for _ in range(i):
    ....:         s += D()
    ....:         n += 1
    ....:
    sage: add_samples(100000)
    sage: while abs(float(s)/n - 0.5) > 5e-2: add_samples(10000)

REFERENCES:

- [DDLL2013]_
"""
#******************************************************************************
#
#                        DGS - Discrete Gaussian Samplers
#
# Copyright (c) 2014, Martin Albrecht  <martinralbrecht+dgs@googlemail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and documentation are
# those of the authors and should not be interpreted as representing official
# policies, either expressed or implied, of the FreeBSD Project.
#*****************************************************************************/

from cysignals.signals cimport sig_on, sig_off

from sage.rings.real_mpfr cimport RealNumber, RealField
from sage.libs.mpfr cimport mpfr_set, MPFR_RNDN
from sage.rings.integer cimport Integer
from sage.misc.randstate cimport randstate, current_randstate

from sage.stats.distributions.dgs cimport dgs_disc_gauss_mp_init, dgs_disc_gauss_mp_clear, dgs_disc_gauss_mp_flush_cache
from sage.stats.distributions.dgs cimport dgs_disc_gauss_dp_init, dgs_disc_gauss_dp_clear, dgs_disc_gauss_dp_flush_cache
from sage.stats.distributions.dgs cimport DGS_DISC_GAUSS_UNIFORM_TABLE, DGS_DISC_GAUSS_UNIFORM_ONLINE, DGS_DISC_GAUSS_UNIFORM_LOGTABLE, DGS_DISC_GAUSS_SIGMA2_LOGTABLE

cdef class DiscreteGaussianDistributionIntegerSampler(SageObject):
    r"""
    A Discrete Gaussian Sampler using rejection sampling.

    .. automethod:: __init__
    .. automethod:: __call__
    """

    # We use tables for σt ≤ table_cutoff
    table_cutoff = 10**6

    def __init__(self, sigma, c=0, tau=6, algorithm=None, precision='mp'):
        r"""
        Construct a new sampler for a discrete Gaussian distribution.

        INPUT:

        - ``sigma`` -- samples `x` are accepted with probability proportional to
          `\exp(-(x-c)^2/(2σ^2))`

        - ``c`` -- the mean of the distribution. The value of ``c`` does not have
          to be an integer. However, some algorithms only support integer-valued
          ``c`` (default: ``0``)

        - ``tau`` -- samples outside the range `(⌊c⌉-⌈στ⌉,...,⌊c⌉+⌈στ⌉)` are
          considered to have probability zero. This bound applies to algorithms which
          sample from the uniform distribution (default: ``6``)

        - ``algorithm`` -- see list below (default: ``'uniform+table'`` for
           `σt` bounded by ``DiscreteGaussianDistributionIntegerSampler.table_cutoff`` and
           ``'uniform+online'`` for bigger `στ`)

        - ``precision`` -- either ``'mp'`` for multi-precision where the actual
          precision used is taken from sigma or ``'dp'`` for double precision. In
          the latter case results are not reproducible. (default: ``'mp'``)

        ALGORITHMS:

        - ``'uniform+table'`` -- classical rejection sampling, sampling from the
          uniform distribution and accepted with probability proportional to
          `\exp(-(x-c)^2/(2σ^2))` where `\exp(-(x-c)^2/(2σ^2))` is precomputed and
          stored in a table. Any real-valued `c` is supported.

        - ``'uniform+logtable'`` -- samples are drawn from a uniform distribution and
          accepted with probability proportional to `\exp(-(x-c)^2/(2σ^2))` where
          `\exp(-(x-c)^2/(2σ^2))` is computed using logarithmically many calls to
          Bernoulli distributions. See [DDLL2013]_ for details.  Only
          integer-valued `c` are supported.

        - ``'uniform+online'`` -- samples are drawn from a uniform distribution and
          accepted with probability proportional to `\exp(-(x-c)^2/(2σ^2))` where
          `\exp(-(x-c)^2/(2σ^2))` is computed in each invocation. Typically this
          is very slow.  See [DDLL2013]_ for details.  Any real-valued `c` is
          accepted.

        - ``'sigma2+logtable'`` -- samples are drawn from an easily samplable
          distribution with `σ = k·σ_2` with `σ_2 = \sqrt{1/(2\log 2)}` and accepted
          with probability proportional to `\exp(-(x-c)^2/(2σ^2))` where
          `\exp(-(x-c)^2/(2σ^2))` is computed using  logarithmically many calls to Bernoulli
          distributions (but no calls to `\exp`). See [DDLL2013]_ for details. Note that this
          sampler adjusts `σ` to match `k·σ_2` for some integer `k`.
          Only integer-valued `c` are supported.

        EXAMPLES::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: DiscreteGaussianDistributionIntegerSampler(3.0, algorithm='uniform+online')
            Discrete Gaussian sampler over the Integers with sigma = 3.000000 and c = 0.000000
            sage: DiscreteGaussianDistributionIntegerSampler(3.0, algorithm='uniform+table')
            Discrete Gaussian sampler over the Integers with sigma = 3.000000 and c = 0.000000
            sage: DiscreteGaussianDistributionIntegerSampler(3.0, algorithm='uniform+logtable')
            Discrete Gaussian sampler over the Integers with sigma = 3.000000 and c = 0.000000

        Note that ``'sigma2+logtable'`` adjusts `σ`::

            sage: DiscreteGaussianDistributionIntegerSampler(3.0, algorithm='sigma2+logtable')
            Discrete Gaussian sampler over the Integers with sigma = 3.397287 and c = 0.000000

        TESTS:

        We are testing invalid inputs::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: DiscreteGaussianDistributionIntegerSampler(-3.0)
            Traceback (most recent call last):
            ...
            ValueError: sigma must be > 0.0 but got -3.000000

            sage: DiscreteGaussianDistributionIntegerSampler(3.0, tau=-1)
            Traceback (most recent call last):
            ...
            ValueError: tau must be >= 1 but got -1

            sage: DiscreteGaussianDistributionIntegerSampler(3.0, tau=2, algorithm='superfastalgorithmyouneverheardof')
            Traceback (most recent call last):
            ...
            ValueError: Algorithm 'superfastalgorithmyouneverheardof' not supported by class 'DiscreteGaussianDistributionIntegerSampler'

            sage: DiscreteGaussianDistributionIntegerSampler(3.0, c=1.5, algorithm='sigma2+logtable')
            Traceback (most recent call last):
            ...
            ValueError: algorithm 'uniform+logtable' requires c%1 == 0

        We are testing correctness for multi-precision::

            sage: def add_samples(i):
            ....:     global mini, maxi, s, n
            ....:     for _ in range(i):
            ....:         x = D()
            ....:         s += x
            ....:         maxi = max(maxi, x)
            ....:         mini = min(mini, x)
            ....:         n += 1

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: D = DiscreteGaussianDistributionIntegerSampler(1.0, c=0, tau=2)
            sage: mini = 1000; maxi = -1000; s = 0; n = 0
            sage: add_samples(2^16)
            sage: while mini != 0 - 2*1.0 or maxi != 0 + 2*1.0 or abs(float(s)/n) >= 0.01:
            ....:     add_samples(2^16)

            sage: D = DiscreteGaussianDistributionIntegerSampler(1.0, c=2.5, tau=2)
            sage: mini = 1000; maxi = -1000; s = 0; n = 0
            sage: add_samples(2^16)
            sage: while mini != 2 - 2*1.0 or maxi != 2 + 2*1.0 or abs(float(s)/n - 2.45) >= 0.01:
            ....:     add_samples(2^16)

            sage: D = DiscreteGaussianDistributionIntegerSampler(1.0, c=2.5, tau=6)
            sage: mini = 1000; maxi = -1000; s = 0; n = 0
            sage: add_samples(2^18)
            sage: while mini > 2 - 4*1.0 or maxi < 2 + 5*1.0 or abs(float(s)/n - 2.5) >= 0.01:  # long time
            ....:     add_samples(2^18)

        We are testing correctness for double precision::

            sage: def add_samples(i):
            ....:     global mini, maxi, s, n
            ....:     for _ in range(i):
            ....:         x = D()
            ....:         s += x
            ....:         maxi = max(maxi, x)
            ....:         mini = min(mini, x)
            ....:         n += 1

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: D = DiscreteGaussianDistributionIntegerSampler(1.0, c=0, tau=2, precision='dp')
            sage: mini = 1000; maxi = -1000; s = 0; n = 0
            sage: add_samples(2^16)
            sage: while mini != 0 - 2*1.0 or maxi != 0 + 2*1.0 or abs(float(s)/n) >= 0.05:
            ....:     add_samples(2^16)

            sage: D = DiscreteGaussianDistributionIntegerSampler(1.0, c=2.5, tau=2, precision='dp')
            sage: mini = 1000; maxi = -1000; s = 0; n = 0
            sage: add_samples(2^16)
            sage: while mini != 2 - 2*1.0 or maxi != 2 + 2*1.0 or abs(float(s)/n - 2.45) >= 0.01:
            ....:     add_samples(2^16)

            sage: D = DiscreteGaussianDistributionIntegerSampler(1.0, c=2.5, tau=6, precision='dp')
            sage: mini = 1000; maxi = -1000; s = 0; n = 0
            sage: add_samples(2^16)
            sage: while mini > -1 or maxi < 6 or abs(float(s)/n - 2.5) >= 0.1:
            ....:     add_samples(2^16)

        We plot a histogram::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: D = DiscreteGaussianDistributionIntegerSampler(17.0)
            sage: S = [D() for _ in range(2^16)]
            sage: list_plot([(v,S.count(v)) for v in set(S)])  # long time
            Graphics object consisting of 1 graphics primitive

        These generators cache random bits for performance reasons. Hence, resetting
        the seed of the PRNG might not have the expected outcome. You can flush this cache with
        ``_flush_cache()``::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: D = DiscreteGaussianDistributionIntegerSampler(3.0)
            sage: sage.misc.randstate.set_random_seed(0); D()
            3
            sage: sage.misc.randstate.set_random_seed(0); D()
            3
            sage: sage.misc.randstate.set_random_seed(0); D._flush_cache(); D()
            3

            sage: D = DiscreteGaussianDistributionIntegerSampler(3.0)
            sage: sage.misc.randstate.set_random_seed(0); D()
            3
            sage: sage.misc.randstate.set_random_seed(0); D()
            3
            sage: sage.misc.randstate.set_random_seed(0); D()
            -3
        """
        if sigma <= 0.0:
            raise ValueError("sigma must be > 0.0 but got %f" % sigma)

        if tau < 1:
            raise ValueError("tau must be >= 1 but got %d" % tau)

        if algorithm is None:
            if sigma*tau <= DiscreteGaussianDistributionIntegerSampler.table_cutoff:
                algorithm = "uniform+table"
            else:
                algorithm = "uniform+online"

        algorithm_str = algorithm

        if algorithm == "uniform+table":
            algorithm = DGS_DISC_GAUSS_UNIFORM_TABLE
        elif algorithm == "uniform+online":
            algorithm = DGS_DISC_GAUSS_UNIFORM_ONLINE
        elif algorithm == "uniform+logtable":
            if (c % 1):
                raise ValueError("algorithm 'uniform+logtable' requires c%1 == 0")
            algorithm = DGS_DISC_GAUSS_UNIFORM_LOGTABLE
        elif algorithm == "sigma2+logtable":
            if (c % 1):
                raise ValueError("algorithm 'uniform+logtable' requires c%1 == 0")
            algorithm = DGS_DISC_GAUSS_SIGMA2_LOGTABLE
        else:
            raise ValueError("Algorithm '%s' not supported by class 'DiscreteGaussianDistributionIntegerSampler'" % (algorithm))

        if precision == "mp":
            if not isinstance(sigma, RealNumber):
                RR = RealField()
                sigma = RR(sigma)

            if not isinstance(c, RealNumber):
                c = sigma.parent()(c)
            sig_on()
            self._gen_mp = dgs_disc_gauss_mp_init((<RealNumber>sigma).value, (<RealNumber>c).value, tau, algorithm)
            sig_off()
            self._gen_dp = NULL
            self.sigma = sigma.parent()(0)
            mpfr_set(self.sigma.value, self._gen_mp.sigma, MPFR_RNDN)
            self.c = c
        elif precision == "dp":
            RR = RealField()
            if not isinstance(sigma, RealNumber):
                sigma = RR(sigma)
            sig_on()
            self._gen_dp = dgs_disc_gauss_dp_init(sigma, c, tau, algorithm)
            sig_off()
            self._gen_mp = NULL
            self.sigma = RR(sigma)
            self.c = RR(c)
        else:
            raise ValueError(f"Parameter precision '{precision}' not supported")

        self.tau = Integer(tau)
        self.algorithm = algorithm_str

    def _flush_cache(self):
        r"""
        Flush the internal cache of random bits.

        EXAMPLES::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler

            sage: f = lambda: sage.misc.randstate.set_random_seed(0)

            sage: f()
            sage: D = DiscreteGaussianDistributionIntegerSampler(30.0)
            sage: [D() for k in range(16)]
            [21, 23, 37, 6, -64, 29, 8, -22, -3, -10, 7, -43, 1, -29, 25, 38]

            sage: f()
            sage: D = DiscreteGaussianDistributionIntegerSampler(30.0)
            sage: l = []
            sage: for i in range(16):
            ....:     f(); l.append(D())
            sage: l
            [21, 21, 21, 21, -21, 21, 21, -21, -21, -21, 21, -21, 21, -21, 21, 21]

            sage: f()
            sage: D = DiscreteGaussianDistributionIntegerSampler(30.0)
            sage: l = []
            sage: for i in range(16):
            ....:     f(); D._flush_cache(); l.append(D())
            sage: l
            [21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21, 21]
        """
        if self._gen_mp:
            dgs_disc_gauss_mp_flush_cache(self._gen_mp)
        if self._gen_dp:
            dgs_disc_gauss_dp_flush_cache(self._gen_dp)

    def __dealloc__(self):
        r"""
        TESTS::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: D = DiscreteGaussianDistributionIntegerSampler(3.0, algorithm='uniform+online')
            sage: del D
        """
        if self._gen_mp:
            dgs_disc_gauss_mp_clear(self._gen_mp)
        if self._gen_dp:
            dgs_disc_gauss_dp_clear(self._gen_dp)

    def __call__(self):
        r"""
        Return a new sample.

        EXAMPLES::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: DiscreteGaussianDistributionIntegerSampler(3.0, algorithm='uniform+online')()  # random
            -3
            sage: DiscreteGaussianDistributionIntegerSampler(3.0, algorithm='uniform+table')()  # random
            3

        TESTS::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: DiscreteGaussianDistributionIntegerSampler(3.0, algorithm='uniform+logtable', precision='dp')() # random output
            13
        """
        cdef randstate rstate
        cdef Integer rop
        if self._gen_mp:
            rstate = current_randstate()
            rop = Integer()
            sig_on()
            self._gen_mp.call(rop.value, self._gen_mp, rstate.gmp_state)
            sig_off()
            return rop
        else:
            sig_on()
            r = self._gen_dp.call(self._gen_dp)
            sig_off()
            return Integer(r)

    def _repr_(self):
        r"""
        TESTS::

            sage: from sage.stats.distributions.discrete_gaussian_integer import DiscreteGaussianDistributionIntegerSampler
            sage: repr(DiscreteGaussianDistributionIntegerSampler(3.0, 2))
            'Discrete Gaussian sampler over the Integers with sigma = 3.000000 and c = 2.000000'
        """
        return f"Discrete Gaussian sampler over the Integers with sigma = {self.sigma:.6f} and c = {self.c:.6f}"



"""
Vectors with rational entries

AUTHOR:

- William Stein (2007)
- Soroosh Yazdani (2007)

EXAMPLES::

    sage: v = vector(QQ,[1,2,3,4,5])
    sage: v
    (1, 2, 3, 4, 5)
    sage: 3*v
    (3, 6, 9, 12, 15)
    sage: v/2
    (1/2, 1, 3/2, 2, 5/2)
    sage: -v
    (-1, -2, -3, -4, -5)
    sage: v - v
    (0, 0, 0, 0, 0)
    sage: v + v
    (2, 4, 6, 8, 10)
    sage: v * v
    55

We make a large zero vector::

    sage: k = QQ^100000; k
    Vector space of dimension 100000 over Rational Field
    sage: v = k(0)
    sage: v[:10]
    (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)

TESTS::

    sage: v = vector(QQ, [1,2/5,-3/8,4])
    sage: loads(dumps(v)) == v
    True

    sage: w = vector(QQ, [-1,0,0,0])
    sage: w.set_immutable()
    sage: isinstance(hash(w), int)
    True
"""

# ****************************************************************************
#       Copyright (C) 2007 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.memory cimport check_allocarray, sig_free
from cysignals.signals cimport sig_on, sig_off

from sage.structure.element cimport Element, Vector
from sage.structure.richcmp cimport rich_to_bool
from sage.rings.integer cimport Integer
from sage.rings.rational cimport Rational

cimport sage.modules.free_module_element as free_module_element

from sage.libs.gmp.mpq cimport *


cdef inline _Rational_from_mpq(mpq_t e):
    cdef Rational z = Rational.__new__(Rational)
    mpq_set(z.value, e)
    return z

cdef class Vector_rational_dense(free_module_element.FreeModuleElement):
    cdef bint is_dense_c(self) noexcept:
        return 1
    cdef bint is_sparse_c(self) noexcept:
        return 0

    def __copy__(self):
        cdef Vector_rational_dense y
        y = self._new_c()
        cdef Py_ssize_t i
        for i in range(self._degree):
            mpq_set(y._entries[i], self._entries[i])
        return y

    cdef int _init(self, Py_ssize_t degree, Parent parent) except -1:
        """
        Initialize the C data structures in this vector. After calling
        this, ``self`` is a zero vector of degree ``degree`` with
        parent ``parent``.

        Only if you call ``__new__`` without a ``parent`` argument, it
        is needed to call this function manually. The only reason to do
        that is for efficiency: calling ``__new__`` without any
        additional arguments besides the type and then calling ``_init``
        is (slightly) more efficient than calling ``__new__`` with a
        ``parent`` argument.

        TESTS:

        Check implicitly that :issue:`10257` works::

            sage: from sage.modules.vector_rational_dense import Vector_rational_dense
            sage: Vector_rational_dense(QQ^(sys.maxsize))
            Traceback (most recent call last):
            ...
            MemoryError: failed to allocate ... bytes
            sage: try:
            ....:     # Note: some malloc() implementations (on OS X
            ....:     # for example) print stuff when an allocation
            ....:     # fails.  # We catch this with the ... in the
            ....:     # doctest result. The * is needed because a
            ....:     # result cannot start with ...
            ....:     print("*")
            ....:     Vector_rational_dense(QQ^(2^56))
            ....: except (MemoryError, OverflowError):
            ....:     print("allocation failed")
            *...
            allocation failed
        """
        # Assign variables only when the array is fully initialized
        cdef mpq_t* entries = <mpq_t*>check_allocarray(degree, sizeof(mpq_t))
        cdef Py_ssize_t i
        sig_on()
        for i in range(degree):
            mpq_init(entries[i])
        sig_off()
        self._entries = entries
        self._degree = degree
        self._parent = parent

    def __cinit__(self, parent=None, x=None, coerce=True, copy=True):
        self._entries = NULL
        self._is_immutable = 0
        if parent is None:
            self._degree = 0
            return
        self._init(parent.degree(), <Parent?>parent)

    def __init__(self, parent, x, coerce=True, copy=True):
        cdef Py_ssize_t i
        cdef Rational z
        if isinstance(x, (list, tuple)):
            if len(x) != self._degree:
                raise TypeError("entries must be a list of length %s" % self._degree)
            for i in range(self._degree):
                z = Rational(x[i])
                mpq_set(self._entries[i], z.value)
            return
        if x != 0:
            raise TypeError("can't initialize vector from nonzero non-list")

    def __dealloc__(self):
        cdef Py_ssize_t i
        if self._entries:
            # Do *not* use sig_on() here, since __dealloc__
            # cannot raise exceptions!
            for i in range(self._degree):
                mpq_clear(self._entries[i])
            sig_free(self._entries)

    cpdef _richcmp_(left, right, int op):
        """
        EXAMPLES::

            sage: v = vector(QQ, [0,0,0,0])
            sage: v == 0
            True
            sage: v == 1
            False
            sage: v == v
            True
            sage: w = vector(QQ, [-1,3/2,0,0])
            sage: w < v
            True
            sage: w > v
            False
            sage: w = vector(QQ, [-1,0,0,0])
            sage: w == w
            True
        """
        cdef Py_ssize_t i
        cdef int c
        for i in range(left.degree()):
            c = mpq_cmp(left._entries[i], (<Vector_rational_dense>right)._entries[i])
            if c < 0:
                return rich_to_bool(op, -1)
            elif c > 0:
                return rich_to_bool(op, 1)
        return rich_to_bool(op, 0)

    cdef get_unsafe(self, Py_ssize_t i):
        """
        EXAMPLES::

            sage: v = vector([1/2,2/3,3/4]); v
            (1/2, 2/3, 3/4)
            sage: v[0]
            1/2
            sage: v[2]
            3/4
            sage: v[-2]
            2/3
            sage: v[5]
            Traceback (most recent call last):
            ...
            IndexError: vector index out of range
            sage: v[-5]
            Traceback (most recent call last):
            ...
            IndexError: vector index out of range
        """
        cdef Rational z = Rational.__new__(Rational)
        mpq_set(z.value, self._entries[i])
        return z

    cdef int set_unsafe(self, Py_ssize_t i, value) except -1:
        """
        EXAMPLES::

            sage: v = vector(QQ, [1/2,2/5,0]); v
            (1/2, 2/5, 0)
            sage: v.set(2, -15/17); v
            (1/2, 2/5, -15/17)
        """
        mpq_set(self._entries[i], (<Rational>value).value)

    def list(self, copy=True):
        """
        The list of entries of the vector.

        INPUT:

        - ``copy`` -- ignored optional argument

        EXAMPLES::

            sage: v = vector(QQ,[1,2,3,4])
            sage: a = v.list(copy=False); a
            [1, 2, 3, 4]
            sage: a[0] = 0
            sage: v
            (1, 2, 3, 4)
        """
        cdef int i
        return [_Rational_from_mpq(self._entries[i])
                for i in range(self._degree)]

    def __reduce__(self):
        return (unpickle_v1, (self._parent, self.list(), self._degree,
                              self._is_immutable))

    cpdef _add_(self, right):
        cdef Vector_rational_dense z, r
        r = right
        z = self._new_c()
        cdef Py_ssize_t i
        for i in range(self._degree):
            mpq_add(z._entries[i], self._entries[i], r._entries[i])
        return z

    cpdef _sub_(self, right):
        cdef Vector_rational_dense z, r
        r = right
        z = self._new_c()
        cdef Py_ssize_t i
        for i in range(self._degree):
            mpq_sub(z._entries[i], self._entries[i], r._entries[i])
        return z

    cpdef _dot_product_(self, Vector right):
        """
        Dot product of dense vectors over the rationals.

        EXAMPLES::

            sage: v = vector(QQ, [1,2,-3]); w = vector(QQ,[4,3,2])
            sage: v*w
            4
            sage: w*v
            4
        """
        cdef Vector_rational_dense r = right
        cdef Rational z
        z = Rational.__new__(Rational)
        cdef mpq_t t
        mpq_init(t)
        mpq_set_si(z.value, 0, 1)
        cdef Py_ssize_t i
        for i in range(self._degree):
            mpq_mul(t, self._entries[i], r._entries[i])
            mpq_add(z.value, z.value, t)
        mpq_clear(t)
        return z

    cpdef _pairwise_product_(self, Vector right):
        """
        EXAMPLES::

            sage: v = vector(QQ, [1,2,-3]); w = vector(QQ,[4,3,2])
            sage: v.pairwise_product(w)
            (4, 6, -6)
        """
        cdef Vector_rational_dense z, r
        r = right
        z = self._new_c()
        cdef Py_ssize_t i
        for i in range(self._degree):
            mpq_mul(z._entries[i], self._entries[i], r._entries[i])
        return z

    cpdef _rmul_(self, Element left):
        cdef Vector_rational_dense z
        cdef Rational a
        if isinstance(left, Rational):
            a = <Rational>left
        elif isinstance(left, Integer):
            a = <Rational>Rational.__new__(Rational)
            mpq_set_z(a.value, (<Integer>left).value)
        else:
            # should not happen
            raise TypeError("cannot convert %s to %s" % (type(left).__name__,
                                                         Rational.__name__))
        z = self._new_c()
        cdef Py_ssize_t i
        for i in range(self._degree):
            mpq_mul(z._entries[i], self._entries[i], a.value)
        return z

    cpdef _lmul_(self, Element right):
        cdef Vector_rational_dense z
        cdef Rational a
        if isinstance(right, Rational):
            a = <Rational>right
        elif isinstance(right, Integer):
            a = <Rational>Rational.__new__(Rational)
            mpq_set_z(a.value, (<Integer>right).value)
        else:
            # should not happen
            raise TypeError("cannot convert %s to %s" % (type(right).__name__,
                                                         Rational.__name__))
        z = self._new_c()
        cdef Py_ssize_t i
        for i in range(self._degree):
            mpq_mul(z._entries[i], self._entries[i], a.value)
        return z

    cpdef _neg_(self):
        cdef Vector_rational_dense z
        z = self._new_c()
        cdef Py_ssize_t i
        for i in range(self._degree):
            mpq_neg(z._entries[i], self._entries[i])
        return z


def unpickle_v0(parent, entries, degree):
    # If you think you want to change this function, don't.
    # Instead make a new version with a name like
    #    make_FreeModuleElement_generic_dense_v1
    # and changed the reduce method below.
    cdef Vector_rational_dense v
    v = Vector_rational_dense.__new__(Vector_rational_dense)
    v._init(degree, parent)
    cdef Rational z
    cdef Py_ssize_t i
    for i in range(degree):
        z = Rational(entries[i])
        mpq_set(v._entries[i], z.value)
    return v


def unpickle_v1(parent, entries, degree, immutable):
    cdef Vector_rational_dense v
    v = Vector_rational_dense.__new__(Vector_rational_dense)
    v._init(degree, parent)
    cdef Rational z
    cdef Py_ssize_t i
    for i in range(degree):
        z = Rational(entries[i])
        mpq_set(v._entries[i], z.value)
    v._is_immutable = immutable
    return v



from __future__ import annotations

from collections.abc import Hashable
from typing import Any, Self
from unittest import TestCase

from sage.categories.category import Category
from sage.typeset.ascii_art import AsciiArt
from sage.typeset.unicode_art import UnicodeArt

_interface_init_with_interface: set[str]

load: Any
loads: Any
dumps: Any
save: Any
unpickle_all: Any
unpickle_global: Any
unpickle_override: Any

CacheKey = Hashable | tuple[Any, ...]

class SageObject:
    _SageObject__custom_name: str | None

    def _test_new(self, **options: Any) -> None: ...
    def rename(self, x: Any | None = None) -> None: ...
    def reset_name(self) -> None: ...
    def get_custom_name(self) -> str | None: ...
    def __repr__(self) -> str: ...
    def _ascii_art_(self) -> AsciiArt: ...
    def _unicode_art_(self) -> UnicodeArt: ...
    def __hash__(self) -> int: ...
    def _cache_key(self) -> CacheKey: ...
    def save(self, filename: str | None = None, compress: bool = True) -> None: ...
    def dump(self, filename: str, compress: bool = True) -> None: ...
    def dumps(self, compress: bool = True) -> bytes: ...
    def category(self) -> Category: ...
    def _test_category(self, **options: Any) -> None: ...
    def parent(self) -> type[Any]: ...
    def _tester(self, **options: Any) -> TestCase: ...
    def _test_not_implemented_methods(self, **options: Any) -> None: ...
    def _test_pickling(self, **options: Any) -> None: ...
    def _sage_(self) -> Self: ...
    def _interface_(self, I: Any) -> Any: ...
    def _interface_init_(self, I: Any | None = None) -> str: ...
    def _interface_is_cached_(self) -> bool: ...
    def _gap_(self, G: Any | None = None) -> Any: ...
    def _gap_init_(self) -> str: ...
    def _libgap_(self) -> Any: ...
    def _gp_(self, G: Any | None = None) -> Any: ...
    def _gp_init_(self) -> str: ...
    def _kash_(self, G: Any | None = None) -> Any: ...
    def _kash_init_(self) -> str: ...
    def _axiom_(self, G: Any | None = None) -> Any: ...
    def _axiom_init_(self) -> str: ...
    def _fricas_(self, G: Any | None = None) -> Any: ...
    def _fricas_init_(self) -> str: ...
    def _giac_(self, G: Any | None = None) -> Any: ...
    def _giac_init_(self) -> str: ...
    def _maxima_(self, G: Any | None = None) -> Any: ...
    def _maxima_init_(self) -> str: ...
    def _maxima_lib_(self, G: Any | None = None) -> Any: ...
    def _maxima_lib_init_(self) -> str: ...
    def _magma_init_(self, magma: Any) -> str: ...
    def _macaulay2_(self, G: Any | None = None) -> Any: ...
    def _macaulay2_init_(self, macaulay2: Any | None = None) -> str: ...
    def _maple_(self, G: Any | None = None) -> Any: ...
    def _maple_init_(self) -> str: ...
    def _mathematica_(self, G: Any | None = None) -> Any: ...
    def _mathematica_init_(self) -> str: ...
    def _mathics_(self, G: Any | None = None) -> Any: ...
    _mathics_init_ = _mathematica_init_
    def _octave_(self, G: Any | None = None) -> Any: ...
    def _octave_init_(self) -> str: ...
    def _polymake_(self, G: Any | None = None) -> Any: ...
    def _polymake_init_(self) -> str: ...
    def _r_init_(self) -> str: ...
    def _singular_(self, G: Any | None = None) -> Any: ...
    def _singular_init_(self) -> str: ...
    def __pari__(self) -> Any: ...
    def _pari_init_(self) -> str: ...
    def _regina_(self, G: Any | None = None) -> Any: ...
    def _regina_init_(self) -> str: ...



# distutils: libraries = flint

include "sage/libs/linkages/padics/fmpz_poly_unram.pxi"
include "sage/libs/linkages/padics/unram_shared.pxi"
include "FP_template.pxi"

cdef class PowComputer_(PowComputer_flint_unram):
    """
    A PowComputer for a floating-point unramified ring or field.
    """
    def __init__(self, Integer prime, long cache_limit, long prec_cap, long ram_prec_cap, bint in_field, poly=None):
        """
        Initialization.

        EXAMPLES::

            sage: R.<a> = ZqFP(125)
            sage: type(R.prime_pow)
            <class 'sage.rings.padics.qadic_flint_FP.PowComputer_'>
            sage: R.prime_pow._prec_type
            'floating-point'
        """
        self._prec_type = 'floating-point'
        PowComputer_flint_unram.__init__(self, prime, cache_limit, prec_cap, ram_prec_cap, in_field, poly)

cdef class qAdicFloatingPointElement(FPElement):
    frobenius = frobenius_unram
    trace = trace_unram
    norm = norm_unram

    def matrix_mod_pn(self):
        r"""
        Return the matrix of right multiplication by the element on
        the power basis `1, x, x^2, \ldots, x^{d-1}` for this
        extension field.  Thus the *rows* of this matrix give the
        images of each of the `x^i`.  The entries of the matrices are
        IntegerMod elements, defined modulo ``p^(self.absprec() / e)``.

        Raises an error if ``self`` has negative valuation.

        EXAMPLES::

            sage: R.<a> = QqFP(5^5,5)
            sage: b = (5 + 15*a)^3
            sage: b.matrix_mod_pn()
            [   125   1125   3375   3375      0]
            [     0    125   1125   3375   3375]
            [380500 377125    125   1125   3375]
            [380500 367000 377125    125   1125]
            [387250 376000 367000 377125    125]

            sage: M = R(0,3).matrix_mod_pn(); M == 0
            True
            sage: M.base_ring()
            Integer Ring

        Check that :issue:`13617` has been fixed::

            sage: R(0).matrix_mod_pn()
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
            [0 0 0 0 0]
        """
        if self.ordp < 0:
            raise ValueError("self must be integral")
        if very_pos_val(self.ordp):
            from sage.matrix.constructor import matrix
            return matrix(ZZ, self.prime_pow.deg, self.prime_pow.deg)
        else:
            return cmatrix_mod_pn(self.unit, self.ordp + self.prime_pow.prec_cap, self.ordp, self.prime_pow)

    def _flint_rep(self, var='x'):
        """
        Replacement for _ntl_rep for use in printing and debugging.

        EXAMPLES::

            sage: R.<a> = QqFP(27, 4)
            sage: (~(1+a))._flint_rep()
            41*x^2 + 40*x + 42
            sage: (1+a)*(41*a^2+40*a+42)
            1
        """
        if self._is_exact_zero():
            raise ValueError("zero does not have a flint rep")
        return self.prime_pow._new_fmpz_poly(self.unit, var)

    def _flint_rep_abs(self, var='x'):
        """
        Replacement for _ntl_rep_abs for use in printing and debugging.

        EXAMPLES::

            sage: R.<a> = QqFP(27, 4)
            sage: (~(3+3*a))._flint_rep_abs()
            (41*x^2 + 40*x + 42, -1)
            sage: (3+3*a)*(41*a^2+40*a+42)
            3
            sage: (3+3*a)._flint_rep_abs()
            (3*x + 3, 0)
        """
        if self.ordp < 0:
            return self._flint_rep(var), Integer(self.ordp)
        if self._is_exact_zero():
            raise ValueError("zero does not have a flint rep")
        cshift_notrunc(self.prime_pow.poly_flint_rep, self.unit, self.ordp, self.ordp + self.prime_pow.prec_cap, self.prime_pow, False)
        return self.prime_pow._new_fmpz_poly(self.prime_pow.poly_flint_rep, var), Integer(0)

    def _modp_rep(self, use_smallest_mode=False, return_list=True):
        r"""
        Return the element with the same reduction mod p that can be expressed
        with coefficients between 0 and p-1.  The absolute precision will be maximal.

        This method is used in printing and computing `p`-adic expansions.

        INPUT:

        - ``use_smallest_mode`` -- if ``True``, use reps between -p/2 and p/2 instead
        - ``return_list`` -- if ``True``, return a list of coefficients (as integers);
          for use in printing

        EXAMPLES::

            sage: R.<a> = Qq(27,4)
            sage: b = a^2 + 5*a - 3
            sage: b._modp_rep()
            ((a^2 + 2*a) + O(3^4), [0, 2, 1])
            sage: b._modp_rep(use_smallest_mode=True)[1]
            [0, -1, 1]
        """
        if self.ordp < 0:
            raise ValueError("No mod p reduction in negative valuation")
        cdef FPElement rep = self._new_c()
        if self.ordp > 0:
            rep._set_exact_zero()
            L = []
        else:
            rep.ordp = 0
            L = cmodp_rep(rep.unit, self.unit, smallest_mode if use_smallest_mode else simple_mode, return_list, self.prime_pow)
        if return_list:
            return rep, L
        else:
            return rep

    def __hash__(self):
        r"""
        Raise a :exc:`TypeError` since this element is not hashable
        (:issue:`11895`).

        TESTS::

            sage: K.<a> = QqFP(9)
            sage: hash(a)
            Traceback (most recent call last):
            ...
            TypeError: unhashable type: 'sage.rings.padics.qadic_flint_FP.qAdicFloatingPointElement'
        """
        # Eventually, hashing will be disabled for all (non-fixed-mod) p-adic
        # elements (#11895), until then, we only to this for types which did
        # not support hashing before we switched some elements to FLINT
        raise TypeError("unhashable type: 'sage.rings.padics.qadic_flint_FP.qAdicFloatingPointElement'")



# sage.doctest: needs sage.rings.finite_rings
"""
Specific algorithms to compute cardinality of elliptic curves over a finite field

Since point counting now uses PARI/GP, this code is only used when a
specific algorithm was specified or when the j-invariant lies in a subfield.

AUTHORS:

- John Cremona (2008-2009): Original point counting code

- Jeroen Demeyer (2017-2018): Refactored and moved to ``cardinality.py``.
"""
# ****************************************************************************
#       Copyright (C) 2008-2009 John Cremona
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from .constructor import EllipticCurve, EllipticCurve_from_j
from sage.schemes.curves.projective_curve import Hasse_bounds
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.rings.finite_rings.finite_field_constructor import FiniteField as GF
from sage.rings.polynomial.polynomial_ring import polygen
from sage.groups.generic import order_from_bounds


def _cardinality_with_j_invariant_1728(self):
    r"""
    Special function to compute cardinality when j=1728.

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.cardinality import _cardinality_with_j_invariant_1728

    An example with q=p=1 (mod 4)::

        sage: F = GF(10009)
        sage: [_cardinality_with_j_invariant_1728(EllipticCurve(F, [0,0,0,11^i,0]))
        ....:  for i in range(4)]
        [10016, 10210, 10004, 9810]

    An example with q=p=3 (mod 4)::

        sage: F = GF(10007)
        sage: [_cardinality_with_j_invariant_1728(EllipticCurve(F, [0,0,0,5^i,0]))
        ....:  for i in range(4)]
        [10008, 10008, 10008, 10008]

    An example with `q=p^2`, p=3 (mod 4)::

        sage: F.<a> = GF(10007^2,'a')
        sage: [_cardinality_with_j_invariant_1728(EllipticCurve(F,[0,0,0,a^i,0]))
        ....:  for i in range(4)]
        [100160064, 100140050, 100120036, 100140050]

    Examples with `q=2^d`, d odd (3 isomorphism classes)::

        sage: F.<a> = GF(2**15,'a')
        sage: ais = [[0,0,1,0,0],[0,0,1,1,0],[0,0,1,1,1]]
        sage: curves = [EllipticCurve(F,ai) for ai in ais]
        sage: all((e1 == e2 or not e1.is_isomorphic(e2)) for e1 in curves for e2 in curves)
        True
        sage: [_cardinality_with_j_invariant_1728(e) for e in curves]
        [32769, 33025, 32513]

    Examples with `q=2^d`, d even (7 isomorphism classes)::

        sage: F.<a> = GF(2**16,'a')
        sage: b = a^11 # trace 1
        sage: ais = [[0,0,1,0,0], [0,0,1,0,b], [0,0,1,b,0], [0,0,a,0,0],
        ....:        [0,0,a,0,a^2*b], [0,0,a^2,0,0], [0,0,a^2,0,a^4*b]]
        sage: curves = [EllipticCurve(F, ai) for ai in ais]
        sage: all((e1 == e2 or not e1.is_isomorphic(e2)) for e1 in curves for e2 in curves)
        True
        sage: [_cardinality_with_j_invariant_1728(e) for e in curves]
        [65025, 66049, 65537, 65793, 65281, 65793, 65281]

    Examples with `q=3^d`, d odd (4 isomorphism classes)::

        sage: F.<a> = GF(3**15,'a')
        sage: b = a^7  # has trace 1
        sage: ais = [[0,0,0,1,0], [0,0,0,-1,0], [0,0,0,-1,b], [0,0,0,-1,-b]]
        sage: curves = [EllipticCurve(F, ai) for ai in ais]
        sage: all((e1 == e2 or not e1.is_isomorphic(e2)) for e1 in curves for e2 in curves)
        True
        sage: [_cardinality_with_j_invariant_1728(e) for e in curves]
        [14348908, 14348908, 14342347, 14355469]

    Examples with `q=3^d`, d even (6 isomorphism classes)::

        sage: F.<g> = GF(3^18,'g')
        sage: i = F(-1).sqrt()
        sage: a = g^8  # has trace 1
        sage: ais = [[0,0,0,1,0], [0,0,0,1,i*a], [0,0,0,g,0],
        ....:        [0,0,0,g^3,0], [0,0,0,g^2,0], [0,0,0,g^2,i*a*g^3]]
        sage: curves = [EllipticCurve(F, ai) for ai in ais]
        sage: all((e1 == e2 or not e1.is_isomorphic(e2)) for e1 in curves for e2 in curves)
        True
        sage: [_cardinality_with_j_invariant_1728(e) for e in curves]
        [387459856, 387400807, 387420490, 387420490, 387381124, 387440173]

    TESTS:

    Check that a bug noted at :issue:`15667` is fixed::

        sage: F.<a> = GF(3^6,'a')
        sage: EllipticCurve([a^5 + 2*a^3 + 2*a^2 + 2*a, a^4 + a^3 + 2*a + 1]).cardinality()
        784
        sage: EllipticCurve([a^5 + 2*a^3 + 2*a^2 + 2*a, a^4 + a^3 + 2*a + 1]).cardinality_exhaustive()
        784
    """
    try:
        return self._order
    except AttributeError:
        pass

    k = self.base_ring()
    assert self.j_invariant() == k(1728)
    q = k.cardinality()
    p = k.characteristic()
    d = k.degree()
    x = polygen(ZZ)

    # p=2, j=0=1728
    #
    # Number of isomorphism classes is 3 (in odd degree) or 7 (in even degree)
    #
    if p == 2:
        if d % 2:
            # The 3 classes are represented, independently of d,
            # by [0,0,1,0,0], [0,0,1,1,0], [0,0,1,1,1]
            E = EllipticCurve(k, [0, 0, 1, 0, 0])
            if self.is_isomorphic(E):
                t = 0
            else:
                n = (d + 1) // 2
                t = 2**n
                n = n % 4
                if n == 0 or n == 1:
                    t = -t
                E = EllipticCurve(k, [0, 0, 1, 1, 1])
                if self.is_isomorphic(E):
                    t = -t
        else:
            # The 7 classes are represented by E1=[0,0,1,0,0],
            # E2=[0,0,1,0,b], E3=[0,0,1,b,0], E4=[0,0,a,0,0],
            # E4=[0,0,a,0,a^2*b], E6=[0,0,a^2,0,0],
            # E7=[0,0,a^2,0,a^4*b], where a is a non-cube and b
            # has trace 1.  E1's Frobenius is pi=(-2)**(d//2); the
            # Frobeniuses are then pi, -pi, 0; w*pi, -w*pi;
            # w^2*pi, -w^2*pi where w is either cube root of
            # unity, so the traces are 2*pi, -2*pi, 0, -pi, +pi;
            # -pi, +pi.
            delta = self.discriminant()
            discube = (delta**((q - 1) // 3) == k(1))
            pi = (-2)**(d // 2)
            if discube:
                a = k.gen()
                b = a
                while b.trace() == 0:
                    b *= a
                if self.is_isomorphic(EllipticCurve(k, [0, 0, 1, b, 0])):
                    t = 0
                else:
                    t = 2 * pi
                    if not self.is_isomorphic(EllipticCurve(k, [0, 0, 1, 0, 0])):
                        t = -t

            else:
                t = pi
                if self.is_isomorphic(EllipticCurve(k, [0, 0, delta, 0, 0])):
                    t = -t

    # p=3, j=0=1728
    #
    # Number of isomorphism classes is 4 (odd degree) or 6 (even degree)
    #
    elif p == 3:
        if d % 2:
            # The 4 classes are represented by [0,0,0,1,0],
            # [0,0,0,-1,0], [0,0,0,-1,a], [0,0,0,-1,-a] where a
            # has trace 1
            delta = self.discriminant()
            if (-delta).is_square():
                t = 0
            else:
                u = delta.sqrt()
                if not u.is_square():
                    u = -u
                tr = ((self.a3()**2 + self.a6()) / u).trace()
                if tr == 0:
                    t = 0
                else:
                    d2 = (d + 1) // 2
                    t = 3**d2
                    if d2 % 2:
                        t = -t
                    if tr == -1:
                        t = -t
        else:
            # The 6 classes are represented by: [0,0,0,1,0],
            # [0,0,0,1,i*a]; [0,0,0,g,0], [0,0,0,g^3,0];
            # [0,0,0,g^2,0], [0,0,0,g^2,i*a*g^3]; where g
            # generates the multiplicative group modulo 4th
            # powers, and a has nonzero trace.

            # The curve is isomorphic to [0,0,0,A4,A6]

            A4 = self.a4() - self.a1() * self.a3()  # = -b4 = 2*b4
            if A4.is_square():
                u = A4.sqrt()
                t = (-3)**(d // 2)
                i = k(-1).sqrt()
                A6 = self.a3()**2 + self.a6()   # = b6
                if (A6 / (i * u * A4)).trace() == 0:
                    t *= 2
                else:
                    t *= -1
                if not u.is_square():
                    t *= -1
            else:
                t = 0

    # p>3, j=1728
    #
    # Number of isomorphism classes is 4 if q=1 (mod 4), else 2
    #
    elif p % 4 == 3:
        if d % 2:
            t = 0
        else:
            t = (-p)**(d // 2)
            w = (self.c4() / k(48))**((q - 1) // 4)
            if w == 1:
                t = 2 * t
            elif w == -1:
                t = -2 * t
            else:
                t = 0

    # p=1 (mod 4).  First find Frobenius pi=a+b*i for [0,0,0,-1,0] over GF(p):
    # N(pi)=p and N(pi-1)=0 (mod 8).
    #
    else:
        R = ZZ.extension(x**2 + 1, 'i')
        i = R.gen(1)
        pi = R.fraction_field().factor(p)[0][0].gens_reduced()[0]
        a, b = pi.list()
        if a % 2 == 0:
            a, b = -b, a
        if (a + b + 1) % 4 == 0:
            a, b = -a, -b
        pi = a + b * i        # Now pi=a+b*i with (a,b)=(1,0),(3,2) mod 4

        # Lift to Frobenius for [0,0,0,-1,0] over GF(p^d):
        if d > 1:
            pi = pi**d
            a, b = pi.list()

        # Compute appropriate quartic twist:
        w = (self.c4() / k(48))**((q - 1) // 4)
        if w == 1:
            t = 2 * a
        elif w == -1:
            t = -2 * a
        elif k(b) == w * k(a):
            t = 2 * b
        else:
            t = -2 * b

    return Integer(q + 1 - t)


def _cardinality_with_j_invariant_0(self):
    r"""
    Special function to compute cardinality when j=0.

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.cardinality import _cardinality_with_j_invariant_0

    An example with q=p=1 (mod 6)::

        sage: F = GF(1009)
        sage: [_cardinality_with_j_invariant_0(EllipticCurve(F,[0,0,0,0,11^i])) for i in range(6)]
        [948, 967, 1029, 1072, 1053, 991]

    An example with q=p=5 (mod 6)::

        sage: F = GF(1013)
        sage: [_cardinality_with_j_invariant_0(EllipticCurve(F,[0,0,0,0,3^i])) for i in range(6)]
        [1014, 1014, 1014, 1014, 1014, 1014]

    An example with `q=p^2`, p=5 (mod 6)::

        sage: F.<a> = GF(1013^2,'a')
        sage: [_cardinality_with_j_invariant_0(EllipticCurve(F,[0,0,0,0,a^i])) for i in range(6)]
        [1028196, 1027183, 1025157, 1024144, 1025157, 1027183]

    For examples in characteristic 2 and 3, see the function
    :func:`_cardinality_with_j_invariant_1728`.
    """
    k = self.base_ring()
    assert self.j_invariant() == k.zero()
    p = k.characteristic()
    if p == 2 or p == 3:  # then 0==1728
        return _cardinality_with_j_invariant_1728(self)

    q = k.cardinality()
    d = k.degree()
    x = polygen(ZZ)

    # p>3, j=0
    #
    # Number of isomorphism classes is 4 if q=1 (mod 4), else 2
    #
    if p % 6 == 5:
        if d % 2:
            t = 0
        else:
            t = (-p)**(d // 2)
            w = (self.c6() / k(-864))**((q - 1) // 6)
            if w == 1:
                t = 2 * t
            elif w == -1:
                t = -2 * t
            elif w**3 == 1:
                t = -t

    # p=1 (mod 6).  First find Frobenius pi=a+b*w for [0,0,0,0,1] over GF(p):
    # N(pi)=p and N(pi-1)=0 (mod 12).
    #
    else:
        R = ZZ.extension(x**2 - x + 1, 'zeta6')
        zeta6 = R.gen(1)
        pi = R.fraction_field().factor(p)[0][0].gens_reduced()[0]
        while (pi - 1).norm() % 12:
            pi *= zeta6
        a, b = pi.list()
        z = k(-b) / k(a)  # a *specific* 6th root of unity in k

        # Now pi=a+b*zeta6 with N(pi-1)=0 (mod 12)

        # Lift to Frobenius for [0,0,0,0,1] over GF(p^d):
        if d > 1:
            pi = pi**d
            a, b = pi.list()

        # Compute appropriate sextic twist:
        w = (self.c6() / k(-864))**((q - 1) // 6)

        if w == 1:
            t = 2 * a + b  # = Trace(pi)
        elif w == -1:
            t = -2 * a - b  # = Trace(-pi)
        elif w == z:
            t = a - b    # = Trace(pi*zeta6)
        elif w == z**2:
            t = -a - 2 * b  # = Trace(pi*zeta6**2)
        elif w == z**4:
            t = b - a    # = Trace(pi*zeta6**4)
        elif w == z**5:
            t = a + 2 * b  # = Trace(pi*zeta6**5)

    return Integer(q + 1 - t)


def cardinality_exhaustive(self):
    r"""
    Return the cardinality of ``self`` over the base field.

    This simply adds up the number of points with each x-coordinate:
    only used for small field sizes!

    EXAMPLES::

        sage: p = next_prime(10^3)
        sage: E = EllipticCurve(GF(p), [3,4])
        sage: E.cardinality_exhaustive()
        1020
        sage: E = EllipticCurve(GF(3^4,'a'), [1,1])
        sage: E.cardinality_exhaustive()
        64
    """
    self._order = Integer(1 + sum(len(self.lift_x(x, all=True))
                                  for x in self.base_field()))
    return self._order


def cardinality_bsgs(self, verbose=False):
    r"""
    Return the cardinality of ``self`` over the base field.

    ALGORITHM: A variant of "Mestre's trick" extended to all finite
    fields by Cremona and Sutherland, 2008.

    .. NOTE::

        1. The Mestre-Schoof-Cremona-Sutherland algorithm may fail for
           a small finite number of curves over `F_q` for `q` at most 49, so
           for `q<50` we use an exhaustive count.

        2. Quadratic twists are not implemented in characteristic 2
           when `j=0 (=1728)`; but this case is treated separately.

    EXAMPLES::

        sage: p = next_prime(10^3)
        sage: E = EllipticCurve(GF(p), [3,4])
        sage: E.cardinality_bsgs()
        1020
        sage: E = EllipticCurve(GF(3^4,'a'), [1,1])
        sage: E.cardinality_bsgs()
        64
        sage: F.<a> = GF(101^3,'a')
        sage: E = EllipticCurve([2*a^2 + 48*a + 27, 89*a^2 + 76*a + 24])
        sage: E.cardinality_bsgs()
        1031352
    """
    E1 = self
    k = self.base_field()
    q = k.order()
    if q < 50:
        if verbose:
            print("q=", q, "< 50 so using exhaustive count")
        return cardinality_exhaustive(self)

    # Construct the quadratic twist:
    E2 = E1.quadratic_twist()
    if verbose:
        print("Quadratic twist is ", E2.ainvs())

    bounds = Hasse_bounds(q)
    lower, upper = bounds
    B = upper - q - 1  # = floor(2*sqrt(q))
    a = ZZ(0)
    N1 = N2 = M = ZZ(1)
    kmin = -B
    kmax = B
    q1 = q + 1
    # Throughout, we have #E=q+1-t where |t|<=B and t=a+k*M = a
    # (mod M) where kmin <= k <= kmax.

    # M is the lcm of the orders of all the points found on E1 and
    # E2, which will eventually exceed 2*B, at which point
    # kmin=kmax.

    if q > 2**10:
        N1 = ZZ(2)**sum([e for P, e in E1._p_primary_torsion_basis(2)])
        N2 = ZZ(2)**sum([e for P, e in E2._p_primary_torsion_basis(2)])
        if q > 2**20:
            N1 *= ZZ(3)**sum([e for P, e in E1._p_primary_torsion_basis(3)])
            N2 *= ZZ(3)**sum([e for P, e in E2._p_primary_torsion_basis(3)])
            if q > 2**40:
                N1 *= ZZ(5)**sum([e for P, e in E1._p_primary_torsion_basis(5)])
                N2 *= ZZ(5)**sum([e for P, e in E2._p_primary_torsion_basis(5)])
        # We now know that t=q+1 (mod N1) and t=-(q+1) (mod N2)
        a = q1
        M = N1
        g, u, v = M.xgcd(N2)  # g==u*M+v*N2
        if N2 > g:
            a = (a * v * N2 - q1 * u * M) // g
            M *= (N2 // g)  # = lcm(M,N2)
            a = a % M
            if verbose:
                print("(a,M)=", (a, M))
            kmin = ((-B - a) / M).ceil()
            kmax = ((B - a) / M).floor()
            if kmin == kmax:
                self._order = q1 - a - kmin * M
                if verbose:
                    print("no random points were needed")
                return self._order
        if verbose:
            print("(2,3,5)-torsion subgroup gives M=", M)

    # N1, N2 are divisors of the orders of E1, E2 separately,
    # which are used to speed up the computation of the orders of
    # points on each curve.  For large q it is worth initializing
    # these with the full order of the (2,3,5)-torsion which are
    # often non-trivial.
    while kmax != kmin:
        # Get a random point on E1 and find its order, using the
        # Hasse bounds and the fact that we know that the group
        # order is a multiple of N1:
        n = order_from_bounds(E1.random_point(), bounds, N1, operation='+')
        if verbose:
            print("New point on E has order ", n)
        # update N1 and M
        N1 = N1.lcm(n)
        g, u, v = M.xgcd(n)  # g==u*M+v*n
        if n > g:
            # update congruence a (mod M) with q+1 (mod n)
            a = (a * v * n + q1 * u * M) // g
            M *= (n // g)  # = lcm(M,n)
            a = a % M
            if verbose:
                print("(a,M)=", (a, M))
            kmin = ((-B - a) / M).ceil()
            kmax = ((B - a) / M).floor()
            if kmin == kmax:
                self._order = q1 - a - kmin * M
                return self._order
            if verbose:
                print("number of possibilities is now ", kmax - kmin + 1)

        # Get a random point on E2 and find its order, using the
        # Hasse bounds and the fact that we know that the group
        # order is a multiple of N2:
        n = order_from_bounds(E2.random_point(), bounds, N2, operation='+')
        if verbose:
            print("New point on E' has order ", n)
        # update N2 and M
        N2 = N2.lcm(n)
        g, u, v = M.xgcd(n)  # g==u*M+v*n
        if n > g:
            # update congruence a (mod M) with -(q+1) (mod n)
            a = (a * v * n - q1 * u * M) // g
            M *= (n // g)  # = lcm(M,n)
            a = a % M
            if verbose:
                print("(a,M)=", (a, M))
            kmin = ((-B - a) / M).ceil()
            kmax = ((B - a) / M).floor()
            if kmin == kmax:
                self._order = q1 - a - kmin * M
                return self._order
            if verbose:
                print("number of possibilities is now ", kmax - kmin + 1)


def _cardinality_subfield(self, jpol):
    """
    Count the number of points on the elliptic curve ``self``, assuming
    that the j-invariant lies in a subfield.

    INPUT:

    - ``jpol`` -- minimal polynomial (over the prime field) of the
      j-invariant of ``self``

    EXAMPLES::

        sage: from sage.schemes.elliptic_curves.cardinality import _cardinality_subfield
        sage: k.<a> = GF(7^5)
        sage: E = EllipticCurve(k, [1,2,3,4,5]); E
        Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 4*x + 5
         over Finite Field in a of size 7^5
        sage: _cardinality_subfield(E, E.j_invariant().minimal_polynomial())
        17019

    TESTS::

        sage: k.<a> = GF(2^6)
        sage: for c in range(500):  # long time
        ....:     ainvs = [k.random_element() for i in range(5)]
        ....:     try:
        ....:         E = EllipticCurve(ainvs)
        ....:     except ArithmeticError:
        ....:         continue  # Singular curve
        ....:     jpol = E.j_invariant().minimal_polynomial()
        ....:     if jpol.degree() < 6:
        ....:         N1 = _cardinality_subfield(E, jpol)
        ....:         N2 = pari.ellcard(E)
        ....:         assert N1 == N2
    """
    k = self.base_ring()
    p = k.characteristic()
    d = k.degree()

    jdeg = jpol.degree()
    if jdeg >= d:
        raise ValueError("j-invariant does not lie in a subfield")

    # we count points on a standard curve with the same
    # j-invariant, defined over the field it generates, then lift
    # to the curve's own field, and finally allow for twists

    # Let j be the j-invariant as element of the smallest finite
    # field over which j is defined.
    GFj = GF((p, jdeg), name='j', modulus=jpol)
    j = GFj.gen()

    # Use special code for j = 0, 1728
    if j == 1728:
        return _cardinality_with_j_invariant_1728(self)
    if j == 0:
        return _cardinality_with_j_invariant_0(self)

    # Recursive call which does all the real work:
    E0 = EllipticCurve_from_j(j)
    N = E0.cardinality(extension_degree=d // jdeg)

    # Map to the original larger field
    phi = GFj.hom([self.j_invariant()])

    # Check if we got a twist. Since j is not 0, 1728 the only twists
    # are quadratic.
    if self.is_isomorphic(E0.base_extend(phi)):
        return N
    q = k.cardinality()
    return 2 * (q + 1) - N



from sage.structure.element cimport MonoidElement, Element
from sage.data_structures.bounded_integer_sequences cimport biseq_t

cdef class QuiverPath(MonoidElement):
    cdef biseq_t _path
    cdef int _start, _end
    cdef QuiverPath _new_(self, int start, int end)
    cpdef _mul_(self, other)
    cpdef _mod_(self, right)
    cpdef tuple complement(self, QuiverPath subpath)
    cpdef bint has_subpath(self, QuiverPath subpath) except -1
    cpdef bint has_prefix(self, QuiverPath subpath) except -1
